from __future__ import annotations

import asyncio
from uuid import uuid4
from enum import Enum, auto
from typing import List, Dict, Any

import ray

from magda.module.module import Module
from magda.pipeline.parallel.group import Group
from magda.pipeline.parallel.future_result import FutureResult


class Job:
    class GroupStatus(Enum):
        IDLE = auto()
        PROCESSING = auto()
        DONE = auto()

    def __init__(self, request: Any, groups: List[Group], is_regular_runtime: bool):
        self.uuid = uuid4()

        self._request = ray.put(request)
        self._groups = groups
        self._is_regular_runtime = is_regular_runtime
        self._output = asyncio.get_running_loop().create_future()
        self._results: Dict[str, Module.Result] = {}
        self._tasks: Dict[str, ray.ObjectRef] = {}
        self._status: Dict[str, Job.GroupStatus] = {
            group.name: Job.GroupStatus.IDLE
            for group in self._groups
        }

    @property
    def status(self) -> Dict[str, Job.GroupStatus]:
        return self._status.copy()

    @property
    def _ready_groups(self) -> List[Group]:
        return [
            group
            for group in self._groups
            if (self._status[group.name] == Job.GroupStatus.IDLE
                and group.fulfills(self._results.keys()))
        ]

    @property
    def _finished(self) -> bool:
        return all([
            self._status[group.name] == Job.GroupStatus.DONE
            for group in self._groups
        ])

    async def _run_group(self, group: Group):
        if self._is_regular_runtime:
            module_dependencies = group.module_dependencies
        else:
            module_dependencies = group.module_dependencies_nonregular

        results = [
            result
            for name, result in self._results.items()
            if name in module_dependencies
        ]
        task = await group.run(self.uuid, self._request, results, self._is_regular_runtime)
        self._tasks[group.name] = task
        self._status[group.name] = Job.GroupStatus.PROCESSING

    async def run(self):
        for group in self._groups:
            dependencies = (
                group.dependencies
                if self._is_regular_runtime
                else group.dependencies_nonregular
            )

            if len(dependencies) == 0:
                await self._run_group(group)

        while not self._finished:
            done, _ = await asyncio.wait(
                fs=self._tasks.values(),
                return_when=asyncio.FIRST_COMPLETED,
            )

            for task in done:
                future: FutureResult = task.result()
                for module_result in future.result:
                    self._results[module_result.name] = module_result
                self._status[future.group] = Job.GroupStatus.DONE
                del self._tasks[future.group]

            for group in self._ready_groups:
                await self._run_group(group)

        return self._results.values()



import os
import pytest

from magda.module.module import Module
from magda.module.factory import ModuleFactory
from magda.module.interface import ModuleInterface
from magda.pipeline.sequential import SequentialPipeline
from magda.config_reader import ConfigReader
from magda.decorators import finalize
from magda.decorators.accepting import accept
from magda.decorators.producing import produce


class MockCorrectInterface(ModuleInterface):
    pass


class MockCorrectDependingInterface(ModuleInterface):
    pass


class MockIncorrectDependingInterface(ModuleInterface):
    pass


class TestInterfaces:

    def setup_method(self, method):
        @produce(MockCorrectDependingInterface)
        @finalize
        class ModuleDependingCorrect(Module.Runtime):
            def run(self, *args, **kwargs):
                pass

        @produce(MockIncorrectDependingInterface)
        @finalize
        class ModuleDependingIncorrect(Module.Runtime):
            def run(self, *args, **kwargs):
                pass

        ModuleFactory.unregister()
        ModuleFactory.register('ModuleDependingCorrect', ModuleDependingCorrect)
        ModuleFactory.register('ModuleDependingIncorrect', ModuleDependingIncorrect)

    def teardown_method(self, method):
        ModuleFactory.unregister()

    def get_config_file(self, config_name):
        __location__ = os.path.realpath(
            os.path.join(os.getcwd(), os.path.dirname(__file__))
        )

        return os.path.join(__location__, 'test_configs', config_name)

    def test_can_build_pipeline_with_module_accepting_interface(self):
        @accept(MockCorrectInterface)
        @finalize
        class ModuleSample(Module.Runtime):
            def run(self, *args, **kwargs):
                pass
        module = ModuleSample('mock')
        builder = SequentialPipeline()
        builder.add_module(module)
        pipeline = builder.build()
        assert isinstance(pipeline, SequentialPipeline.Runtime)

    def test_cannot_register_interfaces(self):
        with pytest.raises(TypeError):
            ModuleFactory.unregister()
            ModuleFactory.register('MockCorrectDependingInterface', MockCorrectDependingInterface)

    def test_can_accept_correct_interface(self):
        @accept(MockCorrectInterface)
        @finalize
        class ModuleSample(Module.Runtime):
            pass

        assert 1 == len(ModuleSample._ancestors)
        assert issubclass(ModuleSample._ancestors[0], MockCorrectInterface)

    def test_cannot_accept_incorrect_interface(self):
        class MockIncorrectInterface:
            pass

        with pytest.raises(TypeError):
            @accept(MockIncorrectInterface)
            @finalize
            class ModuleSample(Module.Runtime):
                pass

    def test_cannot_accept_none_interface(self):
        with pytest.raises(TypeError):
            @accept(None)
            @finalize
            class ModuleSample(Module.Runtime):
                pass

    def test_can_accept_multiple_interfaces(self):
        class MockAdditionalCorrectInterface(ModuleInterface):
            pass

        @accept(MockCorrectInterface, MockAdditionalCorrectInterface)
        @finalize
        class ModuleSample(Module.Runtime):
            pass

        assert 2 == len(ModuleSample._ancestors)
        assert issubclass(ModuleSample._ancestors[0], MockCorrectInterface)
        assert issubclass(ModuleSample._ancestors[1], MockAdditionalCorrectInterface)

    def test_can_accept_modules_and_interfaces(self):
        @finalize
        class ModuleClassSample(Module.Runtime):
            pass

        @accept(MockCorrectInterface, ModuleClassSample)
        @finalize
        class ModuleSample(Module.Runtime):
            pass

        assert 2 == len(ModuleSample._ancestors)
        assert issubclass(ModuleSample._ancestors[0], MockCorrectInterface)
        assert issubclass(ModuleSample._ancestors[1], ModuleClassSample)

    def test_cannot_read_config_with_wrong_all_submodules_interfaces(self):
        @accept(MockCorrectDependingInterface)
        @finalize
        class ModuleSample(Module.Runtime):
            pass

        ModuleFactory.register('ModuleSample', ModuleSample)

        config_file = self.get_config_file('depending_modules_of_invalid_interface.yaml')
        with open(config_file) as config:
            with pytest.raises(Exception):
                pipeline = ConfigReader.read(config, ModuleFactory)

    def test_cannot_read_config_with_wrong_single_submodule_interface(self):
        @accept(MockCorrectDependingInterface)
        @finalize
        class ModuleSample(Module.Runtime):
            pass

        ModuleFactory.register('ModuleSample', ModuleSample)

        config_file = self.get_config_file('depending_modules_partially_invalid_interface.yaml')
        with open(config_file) as config:
            with pytest.raises(Exception):
                pipeline = ConfigReader.read(config, ModuleFactory)

    def test_can_read_config_with_correct_submodule_interfaces(self):
        @accept(MockCorrectDependingInterface)
        @finalize
        class ModuleSample(Module.Runtime):
            def run(self, *args, **kwargs):
                pass

        ModuleFactory.register('ModuleSample', ModuleSample)
        config_file = self.get_config_file('depending_modules_of_correct_interface.yaml')
        with open(config_file) as config:
            pipeline = ConfigReader.read(config, ModuleFactory)

        assert 3 == len(pipeline.modules)
        assert issubclass(pipeline.modules[0].interface, MockCorrectDependingInterface)
        assert issubclass(pipeline.modules[1].interface, MockCorrectDependingInterface)
        assert pipeline.modules[2].input_modules == ['mod1', 'mod2']

    def test_can_produce_correct_interface(self):
        @produce(MockCorrectInterface)
        @finalize
        class ModuleSample(Module.Runtime):
            pass

        assert issubclass(ModuleSample._produce_interface, MockCorrectInterface)

    def test_cannot_produce_multiple_interfaces(self):
        class MockAdditionalCorrectInterface(ModuleInterface):
            pass

        with pytest.raises(TypeError):
            @produce(MockCorrectInterface, MockAdditionalCorrectInterface)
            @finalize
            class ModuleSample(Module.Runtime):
                pass

    def test_cannot_produce_incorrect_interface(self):
        class MockIncorrectInterface:
            pass

        with pytest.raises(TypeError):
            @produce(MockIncorrectInterface)
            @finalize
            class ModuleSample(Module.Runtime):
                pass

    def test_cannot_produce_not_interface(self):
        @finalize
        class ModuleSampleDepending(Module.Runtime):
            pass

        with pytest.raises(TypeError):
            @produce(ModuleSampleDepending)
            @finalize
            class ModuleSample(Module.Runtime):
                pass

    def test_cannot_produce_null(self):
        with pytest.raises(TypeError):
            @produce(None)
            @finalize
            class ModuleSample(Module.Runtime):
                pass

    def test_cannot_produce_empty_interface(self):
        with pytest.raises(TypeError):
            @produce()
            @finalize
            class ModuleSample(Module.Runtime):
                pass



from __future__ import annotations

import asyncio
from uuid import UUID
from typing import List, Iterable

from magda.module.module import Module
from magda.pipeline.parallel.group.actor import Actor
from magda.pipeline.parallel.group.actor_pool import ParallelActorPool
from magda.pipeline.parallel.group.state_type import StateType


class GroupRuntime:
    def __init__(
        self,
        *,
        name: str,
        modules: List[Module],
        dependent_modules: List[Module],
        dependent_modules_nonregular: List[Module],
        replicas: int = 1,
        state_type: StateType = True,
        options={},
    ):
        self.name = name
        self._modules = modules
        self._state_type = state_type
        self.dependencies = set([m.group for m in dependent_modules])
        self.module_dependencies = set([m.name for m in dependent_modules])
        self.dependencies_nonregular = set([m.group for m in dependent_modules_nonregular])
        self.module_dependencies_nonregular = set([m.name for m in dependent_modules_nonregular])
        self.pool = ParallelActorPool([
            Actor.options(**options).remote(name, state_type, modules)
            for _ in range(replicas)
        ])

    @property
    def modules(self) -> List[Module]:
        return self._modules

    @property
    def is_replicated(self) -> bool:
        return self.pool.replicas > 1

    @property
    def state_type(self) -> bool:
        return self._state_type

    def fulfills(self, names: Iterable[str]) -> bool:
        return set(names).issuperset(self.module_dependencies)

    async def run(self, job_id: UUID, request, results, is_regular_runtime):
        return asyncio.ensure_future(
            self.pool.run(
                job_id=job_id,
                request=request,
                results=results,
                is_regular_runtime=is_regular_runtime
            )
        )

    async def teardown(self):
        await self.pool.teardown()



from typing import List

from magda.module.module import Module
from magda.exceptions import DisjointGraphException, CyclicDependenciesException


class GraphValidator:
    def __init__(self, modules: List[Module.Runtime]):
        self.modules = modules
        self.n_modules = len(modules)
        self.neighbours = self.create_neighbours_dict()
        self.successors = self.create_neighbours_dict(successors_only=True)

    def create_neighbours_dict(self, successors_only=False):
        neighbours = {}
        for m in self.modules:
            if m.name not in neighbours:
                neighbours[m.name] = set()
            if not successors_only:
                for input_m in m.input_modules:
                    neighbours[m.name].add(input_m)
            for output_m in m.output_modules:
                neighbours[m.name].add(output_m)
        return neighbours

    def dfs(self, module, visited):
        visited[module] = True

        for neighbour in self.neighbours[module]:
            if not visited[neighbour]:
                self.dfs(neighbour, visited)

    def is_connected(self):
        start_module = self.modules[0].name
        visited = {m.name: False for m in self.modules}
        self.dfs(start_module, visited)
        return all(visited.values())

    def has_cycle(self):
        visited = {m.name: False for m in self.modules}
        start_module = self.modules[0].name
        return self.has_cycle_recursive(start_module, visited, [])

    def has_cycle_recursive(self, module, visited, trace):
        visited[module] = True
        trace.append(module)
        # if any successor is visited and is on recursion stack, there is a cycle
        for neighbour in self.successors[module]:
            if not visited[neighbour]:
                cycle, trace = self.has_cycle_recursive(neighbour, visited, trace)
                if cycle:
                    return True, trace
            elif neighbour in trace:
                trace.append(neighbour)
                return True, trace
        # pop the node from stack
        trace = trace[:-1]
        return False, trace

    def prepare_cycle_string_from_trace(self, trace):
        start_index = trace.index(trace[-1])
        return '->'.join(trace[start_index:])

    def validate(self):
        if len(self.modules) > 1:
            if not self.is_connected():
                raise DisjointGraphException()
            # graph is connected - cycles checked only from start node
            cycle, trace = self.has_cycle()
            if cycle:
                raise CyclicDependenciesException(self.prepare_cycle_string_from_trace(trace))



from __future__ import annotations

from uuid import UUID
from typing import List

import ray

from magda.module.module import Module
from magda.pipeline.graph import Graph
from magda.pipeline.parallel.future_result import FutureResult


@ray.remote
class Actor:
    def __init__(self, name, state_type, modules: List[Module]):
        self.name = name
        self.state_type = state_type
        self.graph = Graph(modules)

    def run(self, job_id: UUID, request, results=[], is_regular_runtime=True):
        return FutureResult(
            job=job_id,
            group=self.name,
            result=self.graph.run(request, results,
                                  is_regular_runtime=is_regular_runtime,
                                  state_type=self.state_type),
        )

    def teardown(self):
        self.graph.teardown()



class CyclicDependenciesException(Exception):
    def __init__(self, trace):
        super().__init__(f'The Graph has a cycle: {trace}')


class DisjointGraphException(Exception):
    def __init__(self):
        super().__init__('The Graph is not connected')


class ClosedPipelineException(Exception):
    def __init__(self):
        super().__init__('The pipeline was closed and cannot be run')



import asyncio
from time import sleep

import pytest
import ray

from magda.pipeline.parallel.group.actor_pool import ParallelActorPool


@ray.remote
class MultiplyActor:
    def __init__(self, factor: int = 2):
        self.factor = factor

    def run(self, data, delay, *args, **kwargs):
        sleep(delay)
        return self.factor * data


@pytest.fixture(scope='module')
def ray_context():
    ray.init(local_mode=True)
    yield None
    ray.shutdown()


class TestActorPool:
    @pytest.mark.parametrize('actors', [1, 2, 5])
    @pytest.mark.asyncio
    async def test_can_create(self, actors, ray_context):
        pool = ParallelActorPool([
            MultiplyActor.remote()
            for _ in range(actors)
        ])
        assert pool.replicas == actors

    @pytest.mark.parametrize('runs,actors', [
        (1, 1),
        (3, 1),
        (1, 2),
        (2, 2),
        (5, 2),
        (1, 5),
        (5, 5),
        (8, 5),
    ])
    @pytest.mark.asyncio
    async def test_should_run(self, runs, actors, ray_context):
        pool = ParallelActorPool([
            MultiplyActor.remote()
            for _ in range(actors)
        ])
        await asyncio.gather(*[
            pool.run(data=0, delay=0.0001)
            for _ in range(runs)
        ])

    @pytest.mark.parametrize('score,runs,factor,actors', [
        (2, 1, 2, 1),
        (9, 3, 3, 1),
        (2, 1, 2, 2),
        (6, 2, 3, 2),
        (15, 5, 3, 2),
        (4, 1, 4, 5),
        (10, 5, 2, 5),
        (24, 8, 3, 5),
    ])
    @pytest.mark.asyncio
    async def test_should_process_correctly(self, score, runs, factor, actors, ray_context):
        pool = ParallelActorPool([
            MultiplyActor.remote(factor)
            for _ in range(actors)
        ])
        results = await asyncio.gather(*[
            pool.run(data=1, delay=0.0001)
            for _ in range(runs)
        ])
        assert sum(results) == score



import pytest

from magda.pipeline.sequential import SequentialPipeline
from magda.module.factory import ModuleFactory
from magda.module.module import Module
from magda.decorators.accepting import accept
from magda.decorators.finalizing import finalize
from magda.exceptions import ClosedPipelineException


@accept(self=True)
@finalize
class MockModule(Module.Runtime):
    def run(self, **kwargs):
        return 'output'


class MockContext:
    pass


class TestSequentialPipeline:
    def setup_method(self, method):
        ModuleFactory.unregister()
        ModuleFactory.register('MockModule', MockModule)
        self.pipeline = SequentialPipeline()
        self.context = MockContext()

    def teardown_method(self, method):
        ModuleFactory.unregister()

    def test_pipeline_add_module(self):
        module = MockModule('MockName')
        self.pipeline.add_module(module)
        assert len(self.pipeline.modules) == 1

    def test_pipeline_add_not_submodule(self):
        class MockNotSubmodule:
            pass

        module = MockNotSubmodule()
        with pytest.raises(TypeError):
            self.pipeline.add_module(module)

    def test_pipeline_get_module_by_name(self):
        module = MockModule('MockName')
        self.pipeline.add_module(module)
        assert self.pipeline.get_module('MockName') == module

    def test_pipeline_get_module_by_name_not_existing(self):
        assert self.pipeline.get_module('MockName') is None

    def test_pipeline_build_no_modules(self):
        runtime = self.pipeline.build(MockContext())
        assert len(runtime.modules) == 0

    def test_pipeline_run_no_modules(self):
        assert len(self.pipeline.modules) == 0
        result = self.pipeline.build(MockContext()).run()
        assert isinstance(result, dict)
        assert len(result) == 0

    def test_pipeline_run_single_module(self):
        module = MockModule('MockName').expose_result()
        self.pipeline.add_module(module)
        results = self.pipeline.build(MockContext()).run()

        assert isinstance(results, dict)
        assert len(results) == 1
        assert 'MockName' in results
        assert results['MockName'] == 'output'

    def test_pipeline_run_multiple_depending_modules(self):
        module_1 = MockModule('module_1')
        module_2 = MockModule('module_2')
        module_2.depends_on(module_1)
        self.pipeline.add_module(module_1)
        self.pipeline.add_module(module_2)
        runtime = self.pipeline.build(MockContext())

        assert len(runtime.modules) == 2
        assert runtime.modules[0].name == 'module_1'
        assert runtime.modules[1].name == 'module_2'

        module_3 = MockModule('module_3')
        module_3.depends_on(module_2)
        self.pipeline.add_module(module_3)
        runtime = self.pipeline.build(MockContext())

        assert len(runtime.modules) == 3
        assert runtime.modules[2].name == 'module_3'

    def test_pipeline_is_regular_flag_shallow(self):
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return 'process'

        @accept(AggregatorModule)
        @finalize
        class RegularModuleAfterAgg(Module.Runtime):
            def run(self, **kwargs):
                return 'run'

        module_aggregator = AggregatorModule('agg_module')
        module_regular = RegularModule('reg_module')
        module_regular_after_agg = RegularModuleAfterAgg('reg_module_after_agg')

        module_aggregator.depends_on(module_regular)
        module_regular_after_agg.depends_on(module_aggregator)

        self.pipeline.add_module(module_regular)
        self.pipeline.add_module(module_aggregator)
        self.pipeline.add_module(module_regular_after_agg)
        runtime = self.pipeline.build(MockContext())

        assert len(runtime.modules) == 3
        assert runtime.modules[0].name == 'reg_module' and runtime.modules[0].is_regular_module
        assert runtime.modules[1].name == 'agg_module' and not runtime.modules[1].is_regular_module
        assert runtime.modules[2].name == 'reg_module_after_agg' and \
            not runtime.modules[2].is_regular_module

    def test_pipeline_is_regular_flag_deep(self):
        @accept(self=True)
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return 'process'

        @accept(AggregatorModule, self=True)
        @finalize
        class RegularModuleAfterAgg(Module.Runtime):
            def run(self, **kwargs):
                return 'run'

        module_aggregator = AggregatorModule('agg_module')
        module_regular1 = RegularModule('reg_module1')
        module_regular11 = RegularModule('reg_module11')
        module_regular12 = RegularModule('reg_module12')
        module_regular_after_agg1 = RegularModuleAfterAgg('reg_module_after_agg1')
        module_regular_after_agg11 = RegularModuleAfterAgg('reg_module_after_agg11')
        module_regular_after_agg12 = RegularModuleAfterAgg('reg_module_after_agg12')
        module_regular_after_agg121 = RegularModuleAfterAgg('reg_module_after_agg121')
        module_regular_after_agg2 = RegularModuleAfterAgg('reg_module_after_agg2')

        module_aggregator.depends_on(module_regular1)
        module_regular11.depends_on(module_regular1)
        module_regular12.depends_on(module_regular1)
        module_regular_after_agg1.depends_on(module_aggregator)
        module_regular_after_agg2.depends_on(module_aggregator)
        module_regular_after_agg11.depends_on(module_regular_after_agg1)
        module_regular_after_agg12.depends_on(module_regular_after_agg1)
        module_regular_after_agg121.depends_on(module_regular_after_agg12)

        self.pipeline.add_module(module_regular1)
        self.pipeline.add_module(module_regular11)
        self.pipeline.add_module(module_regular12)
        self.pipeline.add_module(module_aggregator)
        self.pipeline.add_module(module_regular_after_agg1)
        self.pipeline.add_module(module_regular_after_agg2)
        self.pipeline.add_module(module_regular_after_agg11)
        self.pipeline.add_module(module_regular_after_agg12)
        self.pipeline.add_module(module_regular_after_agg121)
        runtime = self.pipeline.build(MockContext())

        assert len(runtime.modules) == 9
        assert runtime.modules[0].name == 'reg_module1' and runtime.modules[0].is_regular_module
        assert runtime.modules[1].name == 'reg_module12' and runtime.modules[1].is_regular_module
        assert runtime.modules[2].name == 'reg_module11' and runtime.modules[2].is_regular_module
        assert runtime.modules[3].name == 'agg_module' and not runtime.modules[3].is_regular_module
        assert runtime.modules[4].name == 'reg_module_after_agg2' and \
            not runtime.modules[4].is_regular_module
        assert runtime.modules[5].name == 'reg_module_after_agg1' and \
            not runtime.modules[5].is_regular_module
        assert runtime.modules[6].name == 'reg_module_after_agg12' and \
            not runtime.modules[6].is_regular_module
        assert runtime.modules[7].name == 'reg_module_after_agg121' and \
            not runtime.modules[7].is_regular_module
        assert runtime.modules[8].name == 'reg_module_after_agg11' and \
            not runtime.modules[8].is_regular_module

    def test_pipeline_module_depends_on_aggregate_and_regular_error(self):
        @accept(self=True)
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return 'process'

        @accept(AggregatorModule, RegularModule, self=True)
        @finalize
        class RegularModuleFurther(Module.Runtime):
            def run(self, **kwargs):
                return 'run'

        module_regular = RegularModule('1')
        module_aggregator = AggregatorModule('2')
        module_after_agg = RegularModuleFurther('3')
        module_after_regular_and_agg = RegularModuleFurther('4')

        module_after_agg.depends_on(module_aggregator)
        module_after_regular_and_agg.depends_on(module_after_agg)
        module_after_regular_and_agg.depends_on(module_regular)
        module_aggregator.depends_on(module_regular)

        self.pipeline.add_module(module_aggregator)
        self.pipeline.add_module(module_regular)
        self.pipeline.add_module(module_after_regular_and_agg)
        self.pipeline.add_module(module_after_agg)

        with pytest.raises(Exception):
            self.pipeline.build()

    def test_pipeline_modules_run_only_regular(self):
        @accept(self=True)
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'regular run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return 'aggregator run'

        @accept(AggregatorModule, RegularModule, self=True)
        @finalize
        class ModuleAfterAgg(Module.Runtime):
            def run(self, **kwargs):
                return 'after aggregator run'

        module_regular = RegularModule('regular').expose_result()
        module_aggregator = AggregatorModule('aggregator').expose_result()
        module_after_agg = ModuleAfterAgg('after_agg').expose_result()
        module_after_regular = RegularModule('after_reg').expose_result()

        module_aggregator.depends_on(module_regular)
        module_after_agg.depends_on(module_aggregator)
        module_after_regular.depends_on(module_regular)

        self.pipeline.add_module(module_regular)
        self.pipeline.add_module(module_aggregator)
        self.pipeline.add_module(module_after_agg)
        self.pipeline.add_module(module_after_regular)

        results = self.pipeline.build().run()

        assert len(results) == 2
        assert 'regular' in results
        assert 'aggregator' not in results
        assert 'after_agg' not in results
        assert 'after_reg' in results

    def test_pipeline_modules_run_only_aggregate(self):
        @accept(self=True)
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'regular run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return 'aggregator run'

        @accept(AggregatorModule, RegularModule, self=True)
        @finalize
        class ModuleAfterAgg(Module.Runtime):
            def run(self, **kwargs):
                return 'after aggregator run'

        module_regular = RegularModule('regular').expose_result()
        module_aggregator = AggregatorModule('aggregator').expose_result()
        module_after_agg = ModuleAfterAgg('after_agg').expose_result()
        module_after_regular = RegularModule('after_reg').expose_result()

        module_aggregator.depends_on(module_regular)
        module_after_agg.depends_on(module_aggregator)
        module_after_regular.depends_on(module_regular)

        self.pipeline.add_module(module_regular)
        self.pipeline.add_module(module_aggregator)
        self.pipeline.add_module(module_after_agg)
        self.pipeline.add_module(module_after_regular)

        results = self.pipeline.build().process()

        assert len(results) == 2
        assert 'regular' not in results
        assert 'aggregator' in results
        assert 'after_agg' in results
        assert 'after_reg' not in results

    def test_pipeline_module_aggregate_keeps_state(self):
        @accept(self=True)
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'regular run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return self.state

            def aggregate(self, **kwargs):
                self.add_data('state')
                return self.state

        regular = RegularModule('regular').expose_result()
        aggregate = AggregatorModule('aggregate').expose_result()

        aggregate.depends_on(regular)

        self.pipeline.add_module(regular)
        self.pipeline.add_module(aggregate)

        runtime = self.pipeline.build()

        aggregate_runtime_module = runtime.modules[1]
        assert aggregate_runtime_module.name == 'aggregate'
        assert aggregate_runtime_module.state_size == 0
        assert aggregate_runtime_module.state == []

        runtime.run()
        assert aggregate_runtime_module.state_size == 1
        assert aggregate_runtime_module.state == ['state']

        runtime.run()
        assert aggregate_runtime_module.state_size == 2
        assert aggregate_runtime_module.state == ['state', 'state']

    def test_pipeline_module_aggregate_clears_state(self):
        @accept(self=True)
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'regular run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                state_copy = self.state.copy()
                self.clear_state()
                return state_copy

            def aggregate(self, **kwargs):
                self.add_data('state')
                return self.state

        regular = RegularModule('regular').expose_result()
        aggregate = AggregatorModule('aggregate').expose_result()

        aggregate.depends_on(regular)

        self.pipeline.add_module(regular)
        self.pipeline.add_module(aggregate)

        runtime = self.pipeline.build()

        aggregate_runtime_module = runtime.modules[1]
        assert aggregate_runtime_module.name == 'aggregate'

        runtime.run()
        runtime.run()
        assert aggregate_runtime_module.state_size == 2
        assert aggregate_runtime_module.state == ['state', 'state']

        runtime.process()
        assert aggregate_runtime_module.state_size == 0
        assert aggregate_runtime_module.state == []

    def test_pipeline_module_after_gets_aggregated_results(self):
        @accept(self=True)
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'regular run'

        @accept(RegularModule)
        @finalize
        class AggregatorModule(Module.Aggregate):
            def aggregate(self, **kwargs):
                self.add_data('state')
                return self.state

        @accept(AggregatorModule, self=True)
        @finalize
        class ModuleAfterAgg(Module.Runtime):
            def run(self, data, **kwargs):
                return data.of('aggregate')[0] * 2

        regular = RegularModule('regular').expose_result()
        aggregate = AggregatorModule('aggregate').expose_result()
        module_after_agg = ModuleAfterAgg('after_agg').expose_result()

        aggregate.depends_on(regular)
        module_after_agg.depends_on(aggregate)

        self.pipeline.add_module(regular)
        self.pipeline.add_module(aggregate)
        self.pipeline.add_module(module_after_agg)

        runtime = self.pipeline.build()

        aggregate_runtime_module = runtime.modules[1]
        assert aggregate_runtime_module.name == 'aggregate'

        runtime.run()
        runtime.run()
        assert aggregate_runtime_module.state_size == 2
        assert aggregate_runtime_module.state == ['state', 'state']

        results = runtime.process()

        assert 'after_agg' in results
        assert len(results['after_agg']) == 4
        assert results['after_agg'] == ['state', 'state', 'state', 'state']

    def test_pipeline_set_parameters(self):
        parameters = {'param1': 1}
        module = MockModule('MockName').set_parameters(parameters)
        self.pipeline.add_module(module)

        runtime = self.pipeline.build()
        module_runtime = runtime.modules[0]

        assert len(module_runtime.parameters) == 1
        assert 'param1' in module_runtime.parameters

    def test_pipeline_runtime_readonly_parameters(self):
        parameters = {'param1': 1}
        module = MockModule('MockName').set_parameters(parameters)
        self.pipeline.add_module(module)

        runtime = self.pipeline.build()
        module_runtime = runtime.modules[0]

        with pytest.raises(AttributeError):
            module_runtime.parameters = {}

    def test_pipeline_parameters_available_in_run(self):
        @finalize
        class ModuleWithParams(Module.Runtime):
            def run(self, **kwargs):
                return self.parameters

        parameters = {'param1': 1}
        module = ModuleWithParams('module_with_params').expose_result().set_parameters(parameters)

        self.pipeline.add_module(module)
        results = self.pipeline.build().run()

        assert 'module_with_params' in results
        assert results['module_with_params'] == parameters

    def test_pipeline_shared_parameters_available_in_runtime(self):
        shared_parameters = {'shared_param1': 1}
        runtime = self.pipeline.build(shared_parameters=shared_parameters)

        assert len(runtime.shared_parameters) == 1
        assert runtime.shared_parameters == shared_parameters

    def test_pipeline_same_shared_parameters_in_modules(self):
        @finalize
        class ModuleWithParams(Module.Runtime):
            def run(self, **kwargs):
                return self.shared_parameters

        @accept(ModuleWithParams)
        @finalize
        class ModuleWithParamsOther(Module.Runtime):
            def run(self, **kwargs):
                return self.shared_parameters

        module = ModuleWithParams('module_with_params').expose_result()
        module_other = ModuleWithParamsOther('module_with_params_other').expose_result()

        module_other.depends_on(module)
        self.pipeline.add_module(module)
        self.pipeline.add_module(module_other)

        shared_parameters = {'shared_param1': 1}
        results = self.pipeline.build(shared_parameters=shared_parameters).run()

        assert len(results) == 2
        assert 'module_with_params' in results
        assert 'module_with_params_other' in results
        assert results['module_with_params'] == results['module_with_params_other']

    def test_pipeline_build_run_with_defaults(self):
        module = MockModule('mock_name').expose_result()
        self.pipeline.add_module(module)
        results = self.pipeline.build().run()

        assert 'mock_name' in results
        assert results['mock_name'] == 'output'

    def test_pipeline_with_groups(self):
        self.pipeline.add_module(MockModule('m1', group='g1'))
        self.pipeline.add_module(MockModule('m2').depends_on(self.pipeline.get_module('m1')))

        with pytest.warns(UserWarning):
            runtime = self.pipeline.build(MockContext())

    def test_close_pipeline(self):
        @accept(self=True)
        @finalize
        class TestModule(Module.Runtime):
            def bootstrap(self):
                self.state = 'opened'

            def run(self, **kwargs):
                return 'regular run'

            def teardown(self):
                self.state = 'closed'

        module_1 = TestModule('module_1')
        module_2 = TestModule('module_2')
        module_3 = TestModule('module_3')

        module_2.depends_on(module_1)
        module_3.depends_on(module_2)

        self.pipeline.add_module(module_1)
        self.pipeline.add_module(module_2)
        self.pipeline.add_module(module_3)

        runtime = self.pipeline.build()

        runtime.run()
        runtime.close()

        assert runtime._is_closed
        assert runtime.modules[0].state == 'closed'
        assert runtime.modules[1].state == 'closed'
        assert runtime.modules[2].state == 'closed'
        with pytest.raises(ClosedPipelineException):
            runtime.run()
        with pytest.raises(ClosedPipelineException):
            runtime.process()



from __future__ import annotations

from magda.module.runtime import ModuleRuntime
from magda.module.results import ResultSet


class ModuleAggregate(ModuleRuntime):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_state = []

    def process(self, data: ResultSet, **kwargs):
        """ Request Process """
        state_copy = self._current_state.copy()
        self.clear_state()
        return state_copy

    def aggregate(self, data: ResultSet, **kwargs):
        """ Request Aggregate """
        self.add_data(data)
        return self.state

    def clear_state(self):
        self._current_state = []

    def add_data(self, data):
        self._current_state.append(data)

    @property
    def state(self):
        return self._current_state

    @property
    def state_size(self):
        return len(self._current_state)



from __future__ import annotations

from warnings import warn
from typing import Optional

from magda.module.module import Module


class ModuleFactory:
    module_references = {}

    @classmethod
    def register(cls, tag: str, ref):
        if not issubclass(ref, Module):
            raise TypeError('Cannot register not a Module class!')
        if not isinstance(tag, str):
            raise TypeError('Tag must be a string!')
        cls.module_references[tag] = ref

    @classmethod
    def unregister(cls, tag: Optional[str] = None):
        if tag is None:
            cls.module_references.clear()
        elif tag in cls.module_references:
            cls.module_references.pop(tag, None)
        else:
            warn(f"There is no such module in ModuleFactory as {tag}.")

    @classmethod
    def get(cls, tag: str) -> Module:
        return cls.module_references[tag]

    @classmethod
    def create(cls, name: str, module_type: str, module_group: Optional[str] = None):
        if module_type not in cls.module_references:
            raise KeyError(f"'{module_type}' hasn't been registered in the ModuleFactory.")
        return cls.module_references[module_type](name, module_group)



from __future__ import annotations

from typing import Optional
from magda.decorators.common import module_typeguard


def expose(name: Optional[str] = None):
    def wrapper(module_ref):
        module_typeguard(module_ref)
        module_ref._exposed = name or ''
        return module_ref
    return wrapper



import pytest

from magda.module.factory import ModuleFactory
from magda.pipeline.sequential import SequentialPipeline
from magda.module.module import Module
from magda.decorators.accepting import accept
from magda.decorators.finalizing import finalize
from magda.decorators.exposing import expose
from magda.decorators.producing import produce


@expose()
@accept(self=True)
@finalize
class MockModuleContext(Module.Runtime):
    def run(self, **kwargs):
        return self.context


@expose()
@accept(MockModuleContext, self=True)
@finalize
class MockModuleData(Module.Runtime):
    def run(self, data, **kwargs):
        return data


class TestResults:

    def setup_method(self, method):
        ModuleFactory.unregister()
        ModuleFactory.register('MockModuleContext', MockModuleContext)
        ModuleFactory.register('MockModuleData', MockModuleData)
        self.builder = SequentialPipeline()
        self.mod_c = MockModuleContext('mod_c')
        self.mod_d1 = MockModuleData('mod_d1')
        self.mod_d2 = MockModuleData('mod_d2')

    def teardown_method(self, method):
        del self.mod_c
        del self.mod_d1
        del self.mod_d2
        ModuleFactory.unregister()

    def test_string_results(self):
        self.builder.add_module(self.mod_c)
        self.builder.add_module(self.mod_d1)
        self.builder.add_module(self.mod_d2.depends_on(self.mod_c).depends_on(self.mod_d1))
        pipeline = self.builder.build([1, 2, 3])
        runtime = pipeline.run()
        print(runtime)

        assert runtime['mod_d1'].of('mod_c') == []
        assert runtime['mod_d2'].of('mod_c') == [[1, 2, 3]]
        assert runtime['mod_d2'].get('mod_c') == [1, 2, 3]
        assert not runtime['mod_d1'].has('mod_c')
        assert runtime['mod_d2'].has('mod_c')
        assert len(runtime['mod_d1']) == 0
        assert len(runtime['mod_d2']) == 2

    def test_module_results(self):
        self.builder.add_module(self.mod_c)
        self.builder.add_module(self.mod_d1)
        self.builder.add_module(self.mod_d2.depends_on(self.mod_c).depends_on(self.mod_d1))
        pipeline = self.builder.build([1, 2, 3])
        runtime = pipeline.run()

        assert runtime['mod_d1'].of(MockModuleContext) == []
        assert runtime['mod_d2'].of(MockModuleContext) == [[1, 2, 3]]
        assert runtime['mod_d2'].get(MockModuleContext) == [1, 2, 3]
        assert not runtime['mod_d1'].has(MockModuleContext)
        assert runtime['mod_d2'].has(MockModuleContext)

    def test_collection_copy(self):
        self.builder.add_module(self.mod_c)
        self.builder.add_module(self.mod_d1.depends_on(self.mod_c))
        pipeline = self.builder.build([1, 2, 3])
        runtime = pipeline.run()

        assert runtime['mod_d1'].collection
        assert runtime['mod_d1'].collection is not runtime['mod_d1']._collection
        assert runtime['mod_d1'].collection[0] is runtime['mod_d1']._collection[0]

    def test_module_invalid_has(self):
        class InvalidModuleClass():
            pass

        self.builder.add_module(self.mod_c)
        self.builder.add_module(self.mod_d1.depends_on(self.mod_c))
        pipeline = self.builder.build('abc')
        runtime = pipeline.run()

        assert runtime['mod_d1'].has(MockModuleContext)
        with pytest.raises(KeyError):
            assert runtime['mod_d1'].has(InvalidModuleClass)

    def test_module_invalid_filter_of(self):
        class InvalidModuleClass():
            pass

        self.builder.add_module(self.mod_c)
        self.builder.add_module(self.mod_d1.depends_on(self.mod_c))
        pipeline = self.builder.build('abc')
        runtime = pipeline.run()

        assert runtime['mod_d1'].of(MockModuleContext) == ['abc']
        with pytest.raises(KeyError):
            assert runtime['mod_d1'].of(InvalidModuleClass) == []

    def test_module_invalid_get(self):
        self.builder.add_module(self.mod_c)
        self.builder.add_module(self.mod_d1.depends_on(self.mod_c))
        self.builder.add_module(self.mod_d2.depends_on(self.mod_d1))
        self.builder.add_module(
            MockModuleData('mod_d3')
            .depends_on(self.mod_d1)
            .depends_on(self.mod_d2)
        )
        pipeline = self.builder.build('abc')
        runtime = pipeline.run()

        assert runtime['mod_d3'].has(MockModuleData)
        with pytest.raises(Exception):
            assert runtime['mod_d3'].get(MockModuleData)

    def test_module_interface_results(self):
        class MockCorrectInterface(Module.Interface):
            pass

        @expose()
        @produce(MockCorrectInterface)
        @accept(self=True)
        @finalize
        class MockModuleDataInterface(Module.Runtime):
            def run(self, data, **kwargs):
                data.result = 'abc'
                return data

        self.builder.add_module(MockModuleDataInterface('mod_i1'))
        self.builder.add_module(
            MockModuleDataInterface('mod_i2')
            .depends_on(self.builder.get_module('mod_i1'))
        )
        pipeline = self.builder.build()
        runtime = pipeline.run()

        assert not runtime['mod_i1'].has(MockCorrectInterface)
        assert runtime['mod_i2'].has(MockCorrectInterface)
        assert runtime['mod_i2'].get(MockCorrectInterface).result == 'abc'



from __future__ import annotations
from typing import Iterable, Type

from magda.module.module import Module
from magda.decorators.common import module_typeguard


def accept(*ancestors: Iterable[Type[Module]], self=False):
    """ Set module valid ancestors. """
    def wrapper(module_ref):
        module_typeguard(module_ref)

        if not hasattr(module_ref, '_ancestors'):
            module_ref._ancestors = []

        for input_cls in ancestors:
            if not issubclass(input_cls, Module) and not issubclass(input_cls, Module.Interface):
                raise TypeError('The provided argument is not a Module or Module interface class!')
            module_ref._ancestors += [input_cls]

        if self:
            module_ref._ancestors += [module_ref]

        return module_ref

    return wrapper



from time import sleep

from magda.module import Module
from magda.decorators import register, accept, expose, finalize, produce

from examples.modules.common import logger
from examples.interfaces.string import StringInterface


@accept(StringInterface)
@produce(StringInterface)
@expose()
@register('C')
@finalize
class ModuleC(Module.Runtime):
    @logger
    def run(self, data: Module.ResultSet, *args, **kwargs):
        # Access strings (results) from the previous modules
        src = [t.data for t in data.of(StringInterface)]

        # Add delay for example purposes
        #   delay duration is taken from optional parameter `sleep`
        #   and is 2 seconds by default (if the parameter is not provided)
        sleep(self.parameters.get('sleep', 2))

        # Build output string and produce declared interface
        return StringInterface(
            '(' + ' + '.join(src) + (' = ' if len(src) else '') + f'{self.name})'
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Any, Union, Iterable

from magda.module.base import BaseModule, BaseModuleRuntime
from magda.module.interface import ModuleInterface


ResultSetInput = Union[str, BaseModule, BaseModuleRuntime, ModuleInterface]


@dataclass(frozen=True)
class Result:
    """ Module output class """
    result: Any
    interface: ModuleInterface
    name: str
    src_class: BaseModule
    expose: Optional[str] = None


class ResultSet:
    """ Container Type for Module Results

    Basic usage: `has`, `filter`, `of` and `get`. Each of them accepts both string
    (referring to results `expose` name) and `Module` class or it's subclass
    (referring to that module results).

    The direct access to the results list is available via `.collection`.
    """
    def __init__(self, collection: Iterable[Result]):
        self._collection = list(collection)

    @property
    def collection(self) -> List[Result]:
        """ Direct access to underlaying results collection. """
        return self._collection.copy()

    def __len__(self) -> int:
        return len(self._collection)

    def has(self, src: ResultSetInput) -> bool:
        """ Return if results from specific modules are available """
        if isinstance(src, str):
            return any([r.expose == src for r in self._collection])
        elif issubclass(src, ModuleInterface):
            return any([r.interface is src for r in self._collection])
        elif issubclass(src, BaseModule):
            unified_src = src if issubclass(src, BaseModuleRuntime) else src._derived_class
            return any([r.src_class is unified_src for r in self._collection])
        raise KeyError(
            f'Invalid argument: {src}. '
            'Expecting string, Module or Module.Interface class.'
        )

    def filter(self, src: ResultSetInput) -> List[Result]:
        """ List results (with metadata) from specific modules """
        if isinstance(src, str):
            return [r for r in self._collection if r.expose == src]
        elif issubclass(src, ModuleInterface):
            return [r for r in self._collection if r.interface is src]
        elif issubclass(src, BaseModule):
            unified_src = src if issubclass(src, BaseModuleRuntime) else src._derived_class
            return [r for r in self._collection if r.src_class is unified_src]
        raise KeyError(
            f'Invalid argument: {src}. '
            'Expecting string, Module or Module.Interface class.'
        )

    def of(self, src: ResultSetInput) -> List[Any]:
        """ List plain results (without metadata) from specific modules. """
        return [r.result for r in self.filter(src)]

    def get(self, src: ResultSetInput) -> Any:
        """ Return single plain result (without matadata) from specific module.
        Raises Exception when `src` refers to many results.
        """
        results = self.of(src)
        if len(results) > 1:
            raise Exception(
                f'{src} refers to many result. '
                'Expecting selector with only one solution.'
            )
        return results[0]



from __future__ import annotations

import asyncio
from typing import List, Any

import ray

from magda.module.module import Module
from magda.pipeline.base import BasePipeline
from magda.pipeline.parallel.group import Group
from magda.pipeline.parallel.job import Job
from magda.pipeline.parallel.group.state_type import StateType
from magda.exceptions import ClosedPipelineException


class Runtime(BasePipeline.Runtime):
    def __init__(self, groups: List[Group.Runtime], context: Any, shared_parameters: Any):
        super().__init__(context, shared_parameters)
        self.groups = groups
        self._jobs: List[Job] = []
        self._is_closed = False
        self._idle_flag = asyncio.Event()
        self._idle_flag.set()

    @property
    def context(self):
        return ray.get(super().context)

    @property
    def shared_parameters(self):
        return ray.get(super().shared_parameters)

    @property
    def modules(self) -> List[Module.Runtime]:
        return [
            module
            for group in self.groups
            for module in group.modules
        ]

    @property
    def jobs(self) -> List[Job]:
        return self._jobs

    @property
    def closed(self) -> bool:
        return self._is_closed

    async def close(self):
        self._is_closed = True
        await self._idle_flag.wait()
        await asyncio.gather(*[
            group.teardown()
            for group in self.groups
        ])

    def get_group(self, name: str) -> Group:
        return next(iter(filter(lambda group: group.name == name, self.groups)), None)

    def _filter_groups(self, is_regular_runtime=True):
        groups = list(self.groups)
        if is_regular_runtime:
            for group in self.groups:
                if (group.state_type == StateType.AGGREGATE and
                   not any([issubclass(type(mod), Module.Aggregate) for mod in group.modules])):
                    groups.remove(group)
        else:
            for group in self.groups:
                if group.state_type == StateType.REGULAR:
                    groups.remove(group)
        return groups

    async def run(self, request=None, is_regular_runtime=True):
        if self._is_closed:
            raise ClosedPipelineException

        groups = self._filter_groups(is_regular_runtime)
        job = Job(request, groups, is_regular_runtime)

        self._jobs.append(job)
        self._idle_flag.clear()

        results = await job.run()

        self._jobs.remove(job)
        if len(self._jobs) == 0:
            self._idle_flag.set()

        return self.parse_results(results)

    async def process(self, request=None):
        return await self.run(request=request, is_regular_runtime=False)



from time import time
from functools import wraps

from magda.module import Module

from examples.interfaces.common import Context


def log(ref: Module.Runtime, t0: float, msg: str):
    print(f'{ref.__class__.__name__}::{ref.name} [{time() - t0:8.3f}] {msg}')


def logger(fn):
    @wraps(fn)
    def wrapper(ref: Module.Runtime, *args, **kwargs):
        ctx: Context = ref.context
        log(ref, ctx.timer, '- Start')
        output = fn(ref, *args, **kwargs)
        log(ref, ctx.timer, '- End')
        return output
    return wrapper



from __future__ import annotations

from abc import ABC
from typing import Optional


class BaseModule(ABC):
    def __init__(self, name, group: Optional[str] = None):
        self._name = name
        self._group = group

    @property
    def name(self) -> str:
        return self._name

    @property
    def group(self) -> Optional[str]:
        return self._group


class BaseModuleRuntime(BaseModule):
    pass



from __future__ import annotations

from collections import Counter
from abc import ABC, abstractmethod
from typing import List, Dict, Any

from magda.module.module import Module
from magda.pipeline.graph_validator import GraphValidator


class BasePipeline(ABC):
    class Runtime(ABC):
        def __init__(self, context: Any, shared_parameters: Any):
            self._context = context
            self._shared_parameters = shared_parameters

        def parse_results(self, results: List[Module.Result]) -> Dict[str, Any]:
            return {
                r.expose: r.result
                for r in results
                if r.expose is not None
            }

        @property
        def context(self):
            return self._context

        @property
        def shared_parameters(self):
            return self._shared_parameters

        @abstractmethod
        def run(self, request):
            raise NotImplementedError()

        @abstractmethod
        def close(self):
            raise NotImplementedError()

        @property
        @abstractmethod
        def modules(self):
            raise NotImplementedError()

    def __init__(self):
        self.modules: List[Module] = []

    def add_module(self, module) -> BasePipeline:
        if not issubclass(type(module), Module):
            raise TypeError
        self.modules.append(module)
        return self

    def get_module(self, module_name):
        return next((mod for mod in self.modules if mod.name == module_name), None)

    def validate_names(self):
        module_names = [mod.name for mod in self.modules]
        duplicated_modules = [m for m, count in Counter(module_names).items() if count > 1]
        if duplicated_modules:
            raise Exception(
                "Modules must have unique names! "
                f"Found duplicated modules names: {duplicated_modules}"
            )

    def validate_exposition(self):
        exposed = [m.exposed for m in self.modules if m.exposed is not None]
        duplicated_expositions = [ex for ex, count in Counter(exposed).items() if count > 1]
        if duplicated_expositions:
            raise Exception(
                'Each module has to expose results with unique names! '
                f'Found duplicated registration for names: {duplicated_expositions}'
            )

    def validate(self):
        self.validate_names()
        self.validate_exposition()
        GraphValidator(self.modules).validate()

    @abstractmethod
    def build(self, context: Any, shared_parameters: Any) -> BasePipeline.Runtime:
        raise NotImplementedError()



import yaml
from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any

from magda.module.factory import ModuleFactory
from magda.pipeline.sequential import SequentialPipeline
from magda.pipeline.parallel.parallel_pipeline import ParallelPipeline


class ConfigReader:
    @dataclass(frozen=True)
    class ConfigModule:
        name: str
        type: str
        group: Optional[str] = field(default=None)
        depends_on: List[str] = field(default_factory=list)
        parameters: Dict[str, Any] = field(default=None)

    @classmethod
    def read(
        cls,
        config: str,
        module_factory: ModuleFactory,
        context: Optional[Any] = None,
        shared_parameters: Optional[Dict] = None,
    ):
        parsed_yaml = yaml.safe_load(config)
        modules, shared_parameters, group_options = \
            cls._extract_information_from_yaml(parsed_yaml, shared_parameters)

        pipeline = (
            ParallelPipeline()
            if any([m.group is not None for m in modules])
            else SequentialPipeline()
        )

        pipeline = cls._add_modules_to_pipeline(modules, pipeline, module_factory)
        pipeline = cls._add_group_options(group_options, pipeline)

        # connect modules
        for mod in modules:
            curr_mod_obj = pipeline.get_module(mod.name)
            for dependent_mod_name in mod.depends_on:
                dependent_mod_obj = pipeline.get_module(dependent_mod_name)
                if dependent_mod_obj:
                    curr_mod_obj.depends_on(dependent_mod_obj)
                else:
                    raise AttributeError(
                        f"Module '{dependent_mod_name}' hasn't been defined in the config file, "
                        "whereas it's used as a dependency."
                    )
        return pipeline.build(context, shared_parameters)

    @staticmethod
    def _extract_information_from_yaml(parsed_yaml, shared_parameters):
        try:
            modules = [ConfigReader.ConfigModule(**data) for data in parsed_yaml['modules']]
        except TypeError:
            raise Exception("Every module defined in a config file has to have a name and a type.")

        if 'shared_parameters' in parsed_yaml:
            shared_parameters = parsed_yaml['shared_parameters']

        group_options = {}
        if 'groups' in parsed_yaml:
            group_options = {
                group['name']: group['options']
                for group in parsed_yaml['groups']
            }

        return modules, shared_parameters, group_options

    @staticmethod
    def _add_modules_to_pipeline(modules, pipeline, module_factory):
        for mod in modules:
            module = module_factory.create(mod.name, mod.type, mod.group)
            if mod.parameters:
                module.set_parameters(mod.parameters)
            pipeline.add_module(module)
        return pipeline

    @classmethod
    def _add_group_options(cls, group_options, pipeline):
        if group_options and isinstance(pipeline, ParallelPipeline):
            for name, params in group_options.items():
                group = ParallelPipeline.Group(name)
                group.set_replicas(params['replicas'] if 'replicas' in params else 1)
                params.pop('replicas', None)
                if params.keys():
                    group.set_options(**params)
                pipeline.add_group(group)
        return pipeline



from __future__ import annotations

from typing import List, Optional, Any, Dict

from magda.module.base import BaseModule
from magda.module.runtime import ModuleRuntime
from magda.module.interface import ModuleInterface
from magda.module.aggregate import ModuleAggregate
from magda.module.results import Result, ResultSet


class Module(BaseModule):
    """ Generic Module class """
    _ancestors = []
    _exposed: Optional[str] = None
    _produce_interface: Optional[ModuleInterface] = None
    _derived_class: ModuleRuntime = None
    _is_regular_module: bool = True
    _parameters: Dict[str, Any] = {}

    Runtime = ModuleRuntime
    Interface = ModuleInterface
    Aggregate = ModuleAggregate

    Result = Result
    ResultSet = ResultSet

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._input_modules = []
        self._output_modules = []

    @property
    def output_modules(self) -> List[Module]:
        return self._output_modules

    @property
    def input_modules(self) -> List[Module]:
        return self._input_modules

    @property
    def exposed(self) -> str:
        return self.name if self._exposed == '' else self._exposed

    @property
    def is_regular_module(self) -> bool:
        return self._is_regular_module

    @property
    def parameters(self) -> Dict[str, Any]:
        return self._parameters

    @is_regular_module.setter
    def is_regular_module(self, is_regular_module):
        self._is_regular_module = is_regular_module

    def _connect_to(self, module: Module):
        self._output_modules.append(module.name)

    def depends_on(self, module: Module) -> Module:
        if module:
            if module._produce_interface in self._ancestors or module.__class__ in self._ancestors:
                self._input_modules.append(module.name)
                module._connect_to(self)
            else:
                valid_modules = ', '.join([
                    m.Runtime.__name__
                    for m in self._ancestors
                    if issubclass(m, Module)
                ])
                valid_interfaces = ', '.join([
                    i.__name__
                    for i in self._ancestors
                    if issubclass(i, Module.Interface)
                ])
                parent = f'({module.name}: {module._derived_class.__name__})'
                child = f'({self.name}: {self._derived_class.__name__})'
                raise Exception(
                    f'Invalid connection: {parent} -> {child}! Valid ancestors:\n'
                    f' + Modules: [{valid_modules}]\n'
                    f' + Interfaces: [{valid_interfaces}]'
                )
            return self
        else:
            raise TypeError("Dependent module does not exist in the pipeline")

    def set_parameters(self, parameters) -> Module:
        if isinstance(parameters, dict):
            self._parameters = parameters
            return self
        else:
            raise TypeError(f'Parameters must be of type dict but were of type: {type(parameters)}')

    def expose_result(self, name: Optional[str] = None) -> Module:
        self._exposed = name if name is not None else ''
        return self

    def build(self, context=None, shared_parameters=None):
        return self._derived_class(
            name=self.name,
            group=self.group,
            input_modules=self.input_modules,
            output_modules=self.output_modules,
            exposed=self.exposed,
            interface=self._produce_interface,
            context=context,
            shared_parameters=shared_parameters,
            is_regular_module=self.is_regular_module,
            parameters=self._parameters
        )



from __future__ import annotations

from typing import List

from magda.module.module import Module
from magda.pipeline.parallel.group.runtime import GroupRuntime


class Group:
    Runtime = GroupRuntime

    def __init__(
        self, name: str,
        *,
        replicas: int = 1,
        state_type=None,
        **options,
    ):
        self.name = name
        self.replicas = replicas
        self.options = options
        self._state_type = state_type

    def set_replicas(self, replicas: int) -> Group:
        self.replicas = replicas
        return self

    def set_options(self, **options) -> Group:
        self.options = options
        return self

    @property
    def state_type(self):
        return self._state_type

    @state_type.setter
    def state_type(self, state_type):
        self._state_type = state_type

    def build(
        self,
        modules: List[Module],
        dependent_modules: List[Module],
        dependent_modules_nonregular: List[Module],
    ) -> Group.Runtime:
        return self.Runtime(
            name=self.name,
            modules=modules,
            dependent_modules=dependent_modules,
            dependent_modules_nonregular=dependent_modules_nonregular,
            replicas=self.replicas,
            state_type=self.state_type,
            options=self.options,
        )



from __future__ import annotations

from collections import defaultdict
from typing import List

from magda.module.module import Module


class Graph:
    """ Graph class

    Graph holds all modules added to the same group.
    It's responsible for sorting as well as running all modules.
    """

    class TopologicalSorter:
        def __init__(self, modules: List[Module.Runtime]):
            self.modules = modules

        def get(self) -> List[Module.Runtime]:
            """ Get topological order of modules.

            Construct a graph structure as a dictionary, with keys containing nodes
            that are an input to other nodes, and with values equal to nodes they are
            connected to, representing an association list. The graph is then sorted,
            via topological sorting, ensuring an optimal module call sequence.
            """
            scope = [m.name for m in self.modules]
            self.graph = defaultdict(list)
            for mod in self.modules:
                if len(mod.output_modules) > 0:
                    self.graph[mod.name] = [
                        output
                        for output in mod.output_modules
                        if output in scope
                    ]

            visited = {}
            for mod in self.modules:
                visited[mod.name] = False
            stack = []

            for mod in self.modules:
                if not visited[mod.name]:
                    self._topological_sort(mod.name, visited, stack)

            sorted_modules = []
            for s in stack:
                sorted_modules.append(self._get_module(s))
            return sorted_modules

        def _topological_sort(self, v, visited, stack):
            """ Util method for sorting."""
            visited[v] = True
            for i in self.graph[v]:
                if not visited[i]:
                    self._topological_sort(i, visited, stack)
            stack.insert(0, v)

        def _get_module(self, module_name):
            return next((mod for mod in self.modules if mod.name == module_name), None)

    def __init__(self, modules: List[Module.Runtime]):
        self._modules = self.TopologicalSorter(modules).get()
        for module in self._modules:
            module._on_bootstrap()

    @property
    def modules(self) -> List[Module.Runtime]:
        return self._modules.copy()

    def run(self,
            request,
            results: List[Module.Result] = [],
            is_regular_runtime=True,
            state_type=None) -> List[Module.Result]:
        """ Main method for running regular modules in the pipeline

        Modules are run one after another. The output from every module is saved to results array.
        Results are filtered for those containing the inputs for current module,
        and only those are passed to it. The method returns last module's output as it's result.
        """
        results = results.copy()
        for module in self._modules:
            if self._should_be_run(module=module, is_regular_runtime=is_regular_runtime):
                data = Module.ResultSet([r for r in results if r.name in module.input_modules])
                module_result = self._run_method_helper(module, data, request, is_regular_runtime)
                results.append(
                    Module.Result(
                        result=module_result,
                        interface=module.interface,
                        name=module.name,
                        src_class=type(module),
                        expose=self._exposed_result(module, is_regular_runtime),
                    )
                )
        return results

    def teardown(self):
        for module in self._modules:
            module.teardown()

    def _should_be_run(self, module, is_regular_runtime):
        if is_regular_runtime:
            return module.is_regular_module or issubclass(type(module), Module.Aggregate)
        else:
            return not module.is_regular_module or issubclass(type(module), Module.Aggregate)

    def _exposed_result(self, module, is_regular_runtime):
        if is_regular_runtime and issubclass(type(module), Module.Aggregate):
            return None
        else:
            return module.exposed

    def _run_method_helper(self, module, data, request, is_regular_runtime):
        module_result = None
        if issubclass(type(module), Module.Aggregate):
            if is_regular_runtime:
                module.aggregate(data=data, request=request)
                module_result = []
            else:
                module_result = module.process(data=data, request=request)
        else:
            module_result = module.run(data=data, request=request)
        return module_result



from __future__ import annotations

from typing import Optional

import ray

from magda.module.base import BaseModuleRuntime
from magda.module.results import ResultSet, Result


class ModuleRuntime(BaseModuleRuntime):
    def __init__(
        self,
        *,
        name,
        group,
        input_modules,
        output_modules,
        exposed,
        interface,
        parameters,
        context,
        shared_parameters,
        is_regular_module,
    ):
        super().__init__(name, group)
        self._input_modules = input_modules
        self._output_modules = output_modules
        self._exposed = exposed
        self._interface = interface
        self._context = context
        self._shared_parameters = shared_parameters
        self._is_regular_module = is_regular_module
        self._parameters = parameters

    def _on_bootstrap(self):
        if isinstance(self._context, ray.ObjectRef):
            self._context = ray.get(self._context)
        if isinstance(self._shared_parameters, ray.ObjectRef):
            self._shared_parameters = ray.get(self._shared_parameters)
        if callable(self._context):
            self._context = self._context()
        self.bootstrap()

    def bootstrap(self):
        """ Bootstrap module on target device """
        pass

    def teardown(self):
        """ Teardown module on target device """
        pass

    def run(self, data: ResultSet, **kwargs):
        """ Request Run """
        raise NotImplementedError

    @property
    def output_modules(self):
        return self._output_modules

    @property
    def input_modules(self):
        return self._input_modules

    @property
    def exposed(self) -> Optional[str]:
        return self._exposed

    @property
    def interface(self):
        return self._interface

    @property
    def context(self):
        return self._context

    @property
    def is_regular_module(self) -> bool:
        return self._is_regular_module

    @property
    def parameters(self) -> dict:
        return self._parameters

    @property
    def shared_parameters(self) -> dict:
        return self._shared_parameters



import pytest
import ray

from magda.pipeline.parallel import ParallelPipeline
from magda.pipeline.sequential import SequentialPipeline
from magda.module.module import Module
from magda.decorators.accepting import accept
from magda.decorators.finalizing import finalize
from magda.decorators.exposing import expose


@finalize
class MockModule(Module.Runtime):
    def run(self, *args, **kwargs):
        return self.context


@expose()
@accept(MockModule, self=True)
@finalize
class MockModuleReturningContext(Module.Runtime):
    def run(self, *args, **kwargs):
        return self.context


@pytest.fixture(scope='class')
def ray_context():
    ray.init(local_mode=True)
    yield None
    ray.shutdown()


class TestContext:
    """ NOTE that a method _on_bootstrap() from magda/module/runtime.py calls a context
        if it's callable - a function will turn into a value, a class into an object etc.
        That's why callable objects are tested differently here.
        It's to be checked whether this is expected behavior.
    """

    def mock_function(x=-1, y=-1):
        return x + y

    class MockSimpleContext():
        def __init__(self, param=-1):
            self.param = param

    class MockContext():
        def __init__(self, param=-1, obj=None):
            self.param = param
            self.obj = obj

    noncallable_contexts = [
        None,
        dict(key='val'),
        dict(key_str='sth', key_int=123, key_list=[1, 2, 3]),
        list((1, 2, 3, 'a', 'b', 'c')),
        mock_function(1, 2)
    ]

    object_contexts = [
        (MockSimpleContext(param=13), 13),
        (MockContext(param=28), 28),
        (MockContext(param=10, obj=MockSimpleContext()), 10)
    ]

    callable_contexts = [
        (mock_function, int),
        (MockSimpleContext, MockSimpleContext)
    ]

    @pytest.mark.parametrize('context', noncallable_contexts)
    def test_should_accept_noncallable_context(self, context):
        builder = SequentialPipeline()
        m1 = MockModule('m1')
        m2 = MockModuleReturningContext('m2').depends_on(m1)
        builder.add_module(m1)
        builder.add_module(m2)
        pipeline = builder.build(context)

        assert pipeline.context == context
        assert pipeline.modules[0].context == context
        assert pipeline.modules[0].context == pipeline.modules[1].context
        assert pipeline.run()['m2'] == context

    @pytest.mark.parametrize('context', noncallable_contexts)
    @pytest.mark.asyncio
    async def test_parallel_should_accept_noncallable_context(self, context, ray_context):
        builder = ParallelPipeline()
        m1 = MockModule('m1', group='g1')
        m2 = MockModuleReturningContext('m2', group='g2').depends_on(m1)
        builder.add_module(m1)
        builder.add_module(m2)
        pipeline = builder.build(context)

        assert pipeline.context == context
        assert pipeline.modules[0].context == pipeline.modules[1].context
        results = await pipeline.run()
        assert results['m2'] == context

    @pytest.mark.parametrize('context,param', object_contexts)
    def test_object_context(self, context, param):
        builder = SequentialPipeline()
        m1 = MockModule('m1')
        m2 = MockModuleReturningContext('m2').depends_on(m1)
        builder.add_module(m1)
        builder.add_module(m2)
        pipeline = builder.build(context)

        assert pipeline.context == context
        assert pipeline.modules[0].context == context
        assert pipeline.modules[0].context == pipeline.modules[1].context
        assert pipeline.modules[0].context.param == param
        assert pipeline.run()['m2'] == context

    @pytest.mark.parametrize('context,param', object_contexts)
    @pytest.mark.asyncio
    async def test_object_parallel_context(self, context, param, ray_context):
        builder = ParallelPipeline()
        m1 = MockModule('m1', group='g1')
        m2 = MockModuleReturningContext('m2', group='g2').depends_on(m1)
        builder.add_module(m1)
        builder.add_module(m2)
        pipeline = builder.build(context)

        assert pipeline.context.param == param
        results = await pipeline.run()
        assert results['m2'].param == param

    @pytest.mark.parametrize('context,context_type', callable_contexts)
    def test_callable_context(self, context, context_type):
        builder = SequentialPipeline()
        m1 = MockModule('m1')
        m2 = MockModuleReturningContext('m2').depends_on(m1)
        builder.add_module(m1)
        builder.add_module(m2)
        pipeline = builder.build(context)

        assert pipeline.context == context
        assert isinstance(pipeline.modules[0].context, context_type)
        assert isinstance(pipeline.run()['m2'], context_type)

    @pytest.mark.parametrize('context,context_type', callable_contexts)
    @pytest.mark.asyncio
    async def test_callable_parallel_context(self, context, context_type, ray_context):
        builder = ParallelPipeline()
        m1 = MockModule('m1', group='g1')
        m2 = MockModuleReturningContext('m2', group='g2').depends_on(m1)
        builder.add_module(m1)
        builder.add_module(m2)
        pipeline = builder.build(context)

        assert pipeline.context == context
        results = await pipeline.run()
        assert isinstance(results['m2'], context_type)



from __future__ import annotations

from warnings import warn
from typing import List, Dict

from magda.pipeline.base import BasePipeline
from magda.pipeline.graph import Graph
from magda.module.module import Module
from magda.exceptions import ClosedPipelineException


class SequentialPipeline(BasePipeline):
    """ Synchronous processing pipeline """

    class Runtime(BasePipeline.Runtime):
        """ Synchronous pipeline runner """
        def __init__(self, graph: Graph, context=None, shared_parameters=None):
            super().__init__(context, shared_parameters)
            self.graph = graph
            self._is_closed = False

        @property
        def modules(self) -> List[Module]:
            return self.graph.modules

        @property
        def closed(self) -> bool:
            return self._is_closed

        def close(self):
            self._is_closed = True
            self.graph.teardown()

        def run(self, request=None, is_regular_runtime=True) -> Dict[str, Module.Result]:
            if self._is_closed:
                raise ClosedPipelineException()
            results = self.graph.run(request=request, is_regular_runtime=is_regular_runtime)
            return self.parse_results(results)

        def process(self, request=None) -> Dict[str, Module.Result]:
            return self.run(request=request, is_regular_runtime=False)

    def build(self, context=None, shared_parameters=None) -> SequentialPipeline.Runtime:
        self.validate()
        if any([m.group is not None for m in self.modules]):
            msg = ', '.join([
                f'{m.name} ({m.__class__.__name__})'
                for m in self.modules
                if m.group is not None
            ])
            warn(f'The property "group" for modules: ${msg} will be ignored!')

        self._mark_and_validate_modules(modules=self.modules)

        modules = [module.build(context, shared_parameters) for module in self.modules]

        graph = Graph(modules)
        return self.Runtime(graph, context, shared_parameters)

    def _mark_and_validate_modules(self, modules):
        aggregate_modules = self._find_aggregate_modules(modules)
        if len(aggregate_modules) > 0:
            for agg_module in aggregate_modules:
                self._mark_aggregate(agg_module)
            self._validate_module_marking(modules)

    def _find_aggregate_modules(self, modules) -> Module:
        return [m for m in modules if issubclass(m._derived_class, Module.Aggregate)]

    def _mark_aggregate(self, module_start):
        if len(module_start.output_modules) > 0:
            module_output_refs = [
                self.get_module(module_name)
                for module_name in module_start.output_modules
            ]
            for module in module_output_refs:
                module.is_regular_module = False
                self._mark_aggregate(module)

    def _validate_module_marking(self, modules):
        modules_marked_not_runtime = [m for m in modules if not m.is_regular_module]
        for module in modules_marked_not_runtime:
            if len(module.input_modules) >= 2:
                input_modules_flags = [
                    self.get_module(input_module_name).is_regular_module
                    for input_module_name in module.input_modules
                ]
                if False in input_modules_flags and True in input_modules_flags:
                    raise Exception(f'This module, named: {module.name} \
                                    relies on both regular and aggregate modules!')



import pytest

from magda.module.module import Module
from magda.decorators.accepting import accept
from magda.decorators.finalizing import finalize


@finalize
class DependentModule(Module.Runtime):
    def run(self, data, **kwargs):
        pass


@accept(DependentModule)
@finalize
class ConnectedModule(Module.Runtime):
    def run(self, data, **kwargs):
        pass


@finalize
class MockModule(Module.Runtime):
    def run(self, data, **kwargs):
        pass


class TestModule:

    def test_module_init(self):
        connected_module_name = 'testName'
        connected_module = ConnectedModule(connected_module_name)

        assert len(connected_module.input_modules) == 0
        assert len(connected_module.output_modules) == 0
        assert connected_module.name == 'testName'

    def test_module_depends(self):
        connected_module_name = 'connectedName'
        connected_module = ConnectedModule(connected_module_name)

        dependent_module_name = 'dependentName'
        dependent_module = DependentModule(dependent_module_name)

        connected_module.depends_on(dependent_module)

        assert len(connected_module.input_modules) == 1

        input_module_name = connected_module.input_modules[0]
        assert input_module_name == dependent_module.name

    def test_module_is_connected_to(self):
        connected_module_name = 'connectedName'
        connected_module = ConnectedModule(connected_module_name)

        dependent_module_name = 'dependentName'
        dependent_module = DependentModule(dependent_module_name)

        connected_module.depends_on(dependent_module)

        assert len(dependent_module.output_modules) == 1

        output_module_name = dependent_module.output_modules[0]
        assert output_module_name == connected_module.name

    def test_module_concrete_module(self):
        run_output = 'run'
        concrete_name = 'ConcreteModule'

        @finalize
        class ConcreteModule(Module.Runtime):
            def run(self, **kwargs):
                return 'run'
        concrete_module = ConcreteModule('ConcreteModule').build()

        assert concrete_module.name == concrete_name
        assert concrete_module.run() == run_output

    def test_module_not_concrete_module_raises_run_error(self):
        test_name = 'TestName'

        @finalize
        class NotConcreteModule(Module.Runtime):
            pass
        NotConcreteModule._derived_class.__abstractmethods__ = set()
        test_module = NotConcreteModule(test_name).build()
        with pytest.raises(NotImplementedError):
            test_module.run(data=None)

    def test_module_is_aggregator(self):
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return 'process'
        test_module = AggregatorModule('test').build()
        assert not test_module.is_regular_module
        assert issubclass(type(test_module), Module.Aggregate)

    def test_module_is_regular(self):
        @finalize
        class RegularModule(Module.Runtime):
            def run(self, **kwargs):
                return 'run'
        test_module = RegularModule('test').build()
        assert test_module.is_regular_module
        assert \
            issubclass(type(test_module), Module.Runtime) \
            and \
            not issubclass(type(test_module), Module.Aggregate)

    def test_module_aggregator_state(self):
        @finalize
        class AggregatorModule(Module.Aggregate):
            def process(self, **kwargs):
                return 'process'
        test_module = AggregatorModule('test').build()
        assert test_module.state_size == 0
        assert test_module.state == []
        test_module.add_data(1)
        test_module.add_data(2)
        assert test_module.state_size == 2
        assert test_module.state == [1, 2]

    def test_module_parameters(self):
        parameters = {'param1': 1}

        module = MockModule('module_name')
        module.set_parameters(parameters)

        assert isinstance(module.parameters, dict)
        assert 'param1' in module.parameters
        assert module.parameters['param1'] == 1

    def test_module_parameters_default(self):
        module = MockModule('module_name')

        assert module.parameters is not None
        assert isinstance(module.parameters, dict)
        assert len(module.parameters) == 0

    def test_module_parameters_wrong_type_error(self):
        module = MockModule('module_name')

        with pytest.raises(TypeError):
            module.set_parameters([])



from __future__ import annotations

from typing import List, Set

import ray

from magda.module.module import Module
from magda.pipeline.base import BasePipeline
from magda.pipeline.parallel.runtime import Runtime
from magda.pipeline.parallel.group import Group
from magda.pipeline.parallel.group.state_type import StateType


class ParallelPipeline(BasePipeline):
    Runtime = Runtime
    Group = Group

    def __init__(self):
        super().__init__()
        self._groups = []

    def add_group(self, group: Group) -> ParallelPipeline:
        self._groups.append(group)
        return self

    def _list_dependencies(self, module: Module) -> List[Module]:
        return [self.get_module(dep) for dep in module.input_modules]

    def _list_group_dependencies(self, group: str) -> Set[str]:
        return set([
            dep.group
            for m in self.modules
            for dep in self._list_dependencies(m)
            if m.group == group and dep.group != group
        ])

    def build(self, context=None, shared_parameters=None) -> Runtime:
        self.validate()
        if any([m.group is None for m in self.modules]):
            raise Exception('At least one module does not have a group property')

        context_ref = ray.put(context)
        shared_parameters_ref = ray.put(shared_parameters)

        self._mark_and_validate_modules(modules=self.modules)
        runtime_modules = [
            module.build(context_ref, shared_parameters_ref)
            for module in self.modules
        ]

        required_groups = set([m.group for m in runtime_modules])
        defined_groups = set([g.name for g in self._groups])
        missing_groups = required_groups.difference(defined_groups)
        self._groups.extend([
            self.Group(name)
            for name in missing_groups
        ])

        non_pipeline_groups = [g.name for g in self._groups if g.name not in required_groups]
        if non_pipeline_groups:
            raise Exception(
                f"There are groups without any modules: {non_pipeline_groups}. "
                "Please make sure that all groups have at least one module in them."
            )

        self._mark_groups_state(self._groups, runtime_modules)

        runtime_groups = []
        for group in self._groups:
            modules = [m for m in runtime_modules if m.group == group.name]
            dependent_modules_names = [m for module in modules for m in module.input_modules]
            dependent_modules = self._exclude_own_modules(
                [m for m in runtime_modules if m.name in dependent_modules_names],
                modules
            )
            dependent_modules_nonregular = self._exclude_own_modules(
                [m for m in dependent_modules if not m.is_regular_module],
                modules
            )
            runtime_group = group.build(
                modules=modules,
                dependent_modules=dependent_modules,
                dependent_modules_nonregular=dependent_modules_nonregular
            )
            runtime_groups.append(runtime_group)

        return Runtime(runtime_groups, context_ref, shared_parameters_ref)

    @staticmethod
    def _exclude_own_modules(module_dependencies, modules):
        return set(module_dependencies).difference(set([mod for mod in modules]))

    def _mark_groups_state(self, groups, runtime_modules):
        for group in groups:
            group_modules = [m for m in runtime_modules if m.group == group.name]
            if all([m.is_regular_module for m in group_modules]):
                group.state_type = StateType.REGULAR
            elif all([not m.is_regular_module for m in group_modules]):
                group.state_type = StateType.AGGREGATE
            else:
                group.state_type = StateType.MIXED

    def _mark_and_validate_modules(self, modules):
        aggregate_modules = self._find_aggregate_modules(modules)
        if len(aggregate_modules) > 0:
            for agg_module in aggregate_modules:
                self._mark_aggregate(agg_module)
            self._validate_module_marking(modules)

    @staticmethod
    def _find_aggregate_modules(modules) -> Module:
        return [m for m in modules if issubclass(m._derived_class, Module.Aggregate)]

    def _mark_aggregate(self, module_start):
        if len(module_start.output_modules) > 0:
            module_output_refs = [
                self.get_module(module_name)
                for module_name in module_start.output_modules
            ]
            for module in module_output_refs:
                if module is not None:
                    module.is_regular_module = False
                    self._mark_aggregate(module)

    def _validate_module_marking(self, modules):
        modules_marked_not_runtime = [m for m in modules if not m.is_regular_module]
        for module in modules_marked_not_runtime:
            input_modules_flags = [
                self.get_module(input_module_name).is_regular_module
                for input_module_name in module.input_modules
            ]
            if self._has_multiple_items(input_modules_flags):
                raise Exception(f"Module '{module.name}' relies on both "
                                "regular and aggregate modules!")

    @staticmethod
    def _has_multiple_items(input_modules):
        return len(set(input_modules)) > 1



import asyncio
from time import time

from magda.pipeline.parallel import init, ParallelPipeline

from examples.interfaces.common import Context, Request
from examples.modules.a import ModuleA
from examples.modules.b import ModuleB
from examples.modules.c import ModuleC


class ExampleParallelActorPool:
    def __init__(self):
        init()

    @classmethod
    async def demo(cls):
        example = cls()
        await example.build()
        await example.run()

    async def build(self, prefix: str = '{CTX}'):
        builder = ParallelPipeline()
        builder.add_group(ParallelPipeline.Group('g1', replicas=3))
        builder.add_group(ParallelPipeline.Group('g2', replicas=3))
        builder.add_group(ParallelPipeline.Group('g3', replicas=3))

        builder.add_module(ModuleC('m1', group='g1'))
        builder.add_module(ModuleB('m2', group='g2'))
        builder.add_module(
            ModuleB('m3', group='g2')
            .depends_on(builder.get_module('m2'))
        )
        builder.add_module(
            ModuleA('m4', group='g3')
            .depends_on(builder.get_module('m3'))
            .depends_on(builder.get_module('m1'))
            .expose_result('final')
        )

        self.pipeline = builder.build(lambda: Context(prefix))

    async def run(self, value: str = 'R', n_jobs: int = 3):
        # Execute jobs and measure duration
        start = time()
        results = await asyncio.gather(*[
            asyncio.create_task(
                self.pipeline.run(Request(value=f'{value}{i}'))
            )
            for i in range(n_jobs)
        ])
        end = time()

        # Close pipeline
        await self.pipeline.close()
        await asyncio.sleep(0.5)

        # Print results
        print(f'Duration = {end - start:.2f} s')
        print('Results:')
        for index, result in enumerate(results):
            print(f' {index}:')
            for key, value in result.items():
                print(f'  {key}\t → {value}')


if __name__ == "__main__":
    asyncio.run(ExampleParallelActorPool.demo())



from __future__ import annotations
from typing import Type

from magda.module.module import Module
from magda.decorators.common import module_typeguard


def produce(interface_class: Type[Module.Interface]):
    if not issubclass(interface_class, Module.Interface):
        raise TypeError(
            f'Interface {interface_class} '
            'is not subclass of Module.Interface'
        )

    def wrapper(module_ref: Type[Module]):
        module_typeguard(module_ref)
        module_ref._produce_interface = interface_class
        return module_ref

    return wrapper



from __future__ import annotations

from abc import ABCMeta
from magda.module.module import Module


def finalize(module_ref):
    if not issubclass(module_ref, Module.Runtime):
        raise TypeError('Cannot finalize not a Module.Runtime class!')

    if hasattr(module_ref, '_ancestors'):
        ancestors = module_ref._ancestors
        del module_ref._ancestors
    else:
        ancestors = []

    class ModuleBuilderMeta(ABCMeta):
        def __repr__(self) -> str:
            return f'<class {module_ref.__name__}.Builder>'

    class ModuleBuilder(Module, metaclass=ModuleBuilderMeta):
        _ancestors = ancestors
        _derived_class = module_ref
        _is_regular_module = True if not issubclass(module_ref, Module.Aggregate) else False

        def __repr__(self) -> str:
            return f'<{module_ref.__name__}.Builder object>'

    return ModuleBuilder



from dataclasses import dataclass, field
from time import time


@dataclass(frozen=True)
class Request:
    value: str


@dataclass(frozen=True)
class Context:
    prefix: str
    timer: float = field(default_factory=time)



from time import sleep

from magda.module import Module
from magda.decorators import register, accept, finalize, produce

from examples.modules.common import logger, log
from examples.interfaces.common import Request, Context
from examples.interfaces.string import StringInterface
from examples.interfaces.fn import LambdaInterface


@accept(StringInterface, LambdaInterface)
@produce(StringInterface)
@register('A')
@finalize
class ModuleA(Module.Runtime):
    SLEEP_TIME = 2

    def bootstrap(self):
        ctx: Context = self.context
        log(self, ctx.timer, '--- Created!')

    def teardown(self):
        ctx: Context = self.context
        log(self, ctx.timer, '--- Teardown!')

    @logger
    def run(self, data: Module.ResultSet, request: Request):
        # Access context
        ctx: Context = self.context

        # Access results from the previous modules
        #   `src` is a list of strings
        text = [t.fn() for t in data.of(LambdaInterface)]
        src = [t.data for t in data.of(StringInterface)] + text

        # Some heavy computational operations for example
        sleep(self.SLEEP_TIME)

        # Produce declared interface
        return StringInterface(
            f'{ctx.prefix} ('
            + ' + '.join(src)
            + (' = ' if len(src) else '')
            + f'{self.name}) {request.value}'
        )



from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID
from typing import List

from magda.module.module import Module


@dataclass(frozen=True)
class FutureResult:
    job: UUID
    group: str
    result: List[Module.Result]



from time import time

from magda.pipeline.sequential import SequentialPipeline

from examples.interfaces.common import Request, Context
from examples.modules.a import ModuleA
from examples.modules.b import ModuleB
from examples.modules.c import ModuleC


class ExampleSequential:
    def build(self, prefix: str = '{CTX}'):
        builder = SequentialPipeline()
        builder.add_module(ModuleC('m1'))
        builder.add_module(ModuleB('m2'))
        builder.add_module(
            ModuleB('m3')
            .depends_on(builder.get_module('m2'))
        )
        builder.add_module(
            ModuleA('m4')
            .depends_on(builder.get_module('m3'))
            .depends_on(builder.get_module('m1'))
            .expose_result('final')
        )

        self.pipeline = builder.build(lambda: Context(prefix))

    def run(self, value: str = 'R'):
        # Run 1 job and measure duration
        start = time()
        result = self.pipeline.run(Request(value))
        end = time()

        # Close pipeline (and teardown modules)
        self.pipeline.close()

        # Print results
        print(f'Duration = {end - start:.2f} s')
        print('Results:')
        for key, value in result.items():
            print(f'  {key}\t → {value}')


if __name__ == "__main__":
    example = ExampleSequential()
    example.build()
    result = example.run()



from __future__ import annotations

from magda.module.factory import ModuleFactory
from magda.decorators.common import module_typeguard


def register(module_type):
    def wrapper(module_ref):
        module_typeguard(module_ref)
        ModuleFactory.register(module_type, module_ref)
        return module_ref
    return wrapper



import asyncio
from time import sleep
from multiprocessing import Manager

import pytest
import ray

from magda.pipeline.parallel.parallel_pipeline import ParallelPipeline
from magda.module.module import Module
from magda.decorators.accepting import accept
from magda.decorators.finalizing import finalize
from magda.decorators.exposing import expose
from magda.exceptions import ClosedPipelineException


@accept(self=True)
@finalize
class ModuleA(Module.Runtime):
    def run(self, *args, **kwargs):
        sleep(0.03)


@accept(ModuleA, self=True)
@finalize
class ModuleB(Module.Runtime):
    def run(self, *args, **kwargs):
        sleep(0.02)


@accept(ModuleA, ModuleB)
@finalize
class ModuleC(Module.Runtime):
    def run(self, request, *args, **kwargs):
        sleep(0.01)
        return f'{request}:C'


@accept(self=True)
@finalize
class ModuleTestTeardown(Module.Runtime):
    def bootstrap(self):
        self.context['ns'].counter += 1

    def run(self, **kwargs):
        return 'run'

    def teardown(self):
        self.context['ns'].counter -= 1


@pytest.fixture(scope='class')
def ray_context():
    ray.init(local_mode=True)
    yield None
    ray.shutdown()


class ModuleParameters:
    @staticmethod
    def single_module_group():
        return [ModuleB('m1', group='g1')]

    @staticmethod
    def two_modules_one_groups():
        m1 = ModuleA('m1', group='g1')
        m2 = ModuleB('m2', group='g1').depends_on(m1)
        return [m1, m2]

    @staticmethod
    def two_modules_two_groups():
        m1 = ModuleA('m1', group='g1')
        m2 = ModuleB('m2', group='g2').depends_on(m1)
        return [m1, m2]

    @staticmethod
    def four_modules_two_groups():
        m1 = ModuleA('m1', group='g1')
        m2 = ModuleA('m2', group='g1').depends_on(m1)
        m3 = ModuleB('m3', group='g2').depends_on(m2)
        m4 = ModuleC('m4', group='g2').depends_on(m3)
        return [m1, m2, m3, m4]


class TestParallelPipelineSerial:

    @pytest.mark.asyncio
    async def test_can_create(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(ModuleA('m1', group='g1'))
        pipeline = builder.build()
        assert isinstance(pipeline, ParallelPipeline.Runtime)
        assert len(pipeline.groups) == 1
        assert set([g.name for g in pipeline.groups]) == set(['g1'])

    @pytest.mark.asyncio
    async def test_can_create_multiple_groups(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(ModuleA('m1', group='g1'))
        builder.add_module(ModuleB('m2', group='g2').depends_on(builder.get_module('m1')))
        pipeline = builder.build()
        assert isinstance(pipeline, ParallelPipeline.Runtime)
        assert len(pipeline.groups) == 2
        assert set([g.name for g in pipeline.groups]) == set(['g1', 'g2'])

    @pytest.mark.asyncio
    async def test_can_create_multiple_dependent_groups(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(ModuleA('m1', group='g1'))
        builder.add_module(ModuleB('m2', group='g2').depends_on(builder.get_module('m1')))
        builder.add_module(ModuleC('m3', group='g3').depends_on(builder.get_module('m2')))
        pipeline = builder.build()
        assert isinstance(pipeline, ParallelPipeline.Runtime)
        assert len(pipeline.groups) == 3
        assert set([g.name for g in pipeline.groups]) == set(['g1', 'g2', 'g3'])
        g1 = next((g for g in pipeline.groups if g.name == 'g1'))
        g2 = next((g for g in pipeline.groups if g.name == 'g2'))
        g3 = next((g for g in pipeline.groups if g.name == 'g3'))
        assert g1.dependencies == set([])
        assert g2.dependencies == set(['g1'])
        assert g3.dependencies == set(['g2'])

    @pytest.mark.asyncio
    async def test_cannot_create_empty_group(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(ModuleA('m1', group='g1'))
        builder.add_module(ModuleB('m2', group='g2').depends_on(builder.get_module('m1')))
        builder.add_module(ModuleC('m3', group='g3').depends_on(builder.get_module('m2')))
        builder.add_group(builder.Group('g4'))
        with pytest.raises(Exception):
            pipeline = builder.build()

    @pytest.mark.parametrize('modules', [
        [ModuleB('m2')],
        [ModuleB('m2', group='g1'), ModuleB('m2')],
        [ModuleB('m2'), ModuleB('m2', group='g1')],
        [ModuleB('m2'), ModuleB('m2', group='g1'), ModuleB('m2')],
        [ModuleB('m2'), ModuleB('m2', group='g1'), ModuleB('m2', group='g1')],
    ])
    @pytest.mark.asyncio
    async def test_cannot_create_without_group(self, modules, ray_context):
        builder = ParallelPipeline()
        for module in modules:
            builder.add_module(module)
        with pytest.raises(Exception):
            pipeline = builder.build()

    @pytest.mark.parametrize('modules', [
        ModuleParameters.single_module_group(),
        ModuleParameters.two_modules_one_groups(),
        ModuleParameters.two_modules_two_groups(),
        ModuleParameters.four_modules_two_groups()
    ])
    @pytest.mark.asyncio
    async def test_listing_group_modules(self, modules, ray_context):
        builder = ParallelPipeline()
        for module in modules:
            builder.add_module(module)
        pipeline = builder.build()

        names = set([m.name for m in modules])
        pipeline_names = set([m.name for m in pipeline.modules])
        assert names == pipeline_names

        def prepare_comparison(group) -> bool:
            expected = set([
                m.name
                for m in modules
                if m.group == group.name
            ])
            current = set([m.name for m in group.modules])
            return expected == current

        compared_modules = [prepare_comparison(group) for group in pipeline.groups]
        assert all(compared_modules)

    @pytest.mark.asyncio
    async def test_can_pass_context(self, ray_context):
        context = dict(ctx='MAGDA')
        tag = 'output'

        @expose(tag)
        @finalize
        class ModuleWithContext(Module.Runtime):
            def run(self, *args, **kwargs):
                return self.context

        builder = ParallelPipeline()
        builder.add_module(ModuleWithContext('m1', group='g1'))
        pipeline = builder.build(context)
        assert pipeline.context == context
        results = await pipeline.run()
        assert results[tag] == context

    @pytest.mark.asyncio
    async def test_can_pass_shared_parameters(self, ray_context):
        context = dict(ctx='MAGDA')
        shared_parameters = {'shared_param1': 1}
        tag = 'output'

        @expose(tag)
        @finalize
        class ModuleWithSharedParams(Module.Runtime):
            def run(self, *args, **kwargs):
                return self.shared_parameters

        builder = ParallelPipeline()
        builder.add_module(ModuleWithSharedParams('m1', group='g1'))
        pipeline = builder.build(context, shared_parameters)
        assert pipeline.shared_parameters == shared_parameters
        results = await pipeline.run()
        assert results[tag] == shared_parameters

    @pytest.mark.asyncio
    async def test_build_run_with_defaults(self, ray_context):
        tag = 'output_tag'

        @expose(tag)
        @finalize
        class MockModule(Module.Runtime):
            def run(self, *args, **kwargs):
                sleep(0.01)
                return 'output_result'

        builder = ParallelPipeline()
        builder.add_module(MockModule('m1', group='g1'))
        results = await builder.build().run()

        assert 'output_tag' in results
        assert results['output_tag'] == 'output_result'


class TestParallelPipeline:
    @pytest.mark.asyncio
    async def test_request(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(ModuleA('m1', group='g1'))
        builder.add_module(ModuleB('m2', group='g2'))
        builder.add_module(
            ModuleB('m3', group='g2')
            .depends_on(builder.get_module('m2'))
        )
        builder.add_module(
            ModuleC('m4', group='g3')
            .depends_on(builder.get_module('m3'))
            .depends_on(builder.get_module('m1'))
            .expose_result('final')
        )
        pipeline = builder.build()
        result = await pipeline.run('R1')
        assert result['final'] == 'R1:C'

    @pytest.mark.asyncio
    async def test_request_checking_dependencies(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(ModuleA('m0', group='g0'))
        builder.add_module(ModuleA('m1', group='g1'))
        builder.add_module(
            ModuleB('m2', group='g2')
            .depends_on(builder.get_module('m0'))
        )
        builder.add_module(
            ModuleB('m3', group='g2')
            .depends_on(builder.get_module('m2'))
        )
        builder.add_module(
            ModuleC('m4', group='g3')
            .depends_on(builder.get_module('m3'))
            .depends_on(builder.get_module('m1'))
            .expose_result('final')
        )
        pipeline = builder.build()
        result = await pipeline.run('R1')
        assert result['final'] == 'R1:C'

    @pytest.mark.asyncio
    async def test_parallel_requests(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(ModuleA('m1', group='g1'))
        builder.add_module(ModuleB('m2', group='g2'))
        builder.add_module(
            ModuleB('m3', group='g2')
            .depends_on(builder.get_module('m2'))
        )
        builder.add_module(
            ModuleC('m4', group='g3')
            .depends_on(builder.get_module('m3'))
            .depends_on(builder.get_module('m1'))
            .expose_result('final')
        )
        pipeline = builder.build()
        results = await asyncio.gather(
            asyncio.create_task(pipeline.run('R1')),
            asyncio.create_task(pipeline.run('R2')),
            asyncio.create_task(pipeline.run('R3')),
        )
        outputs = set([r['final'] for r in results])
        assert outputs == set(['R1:C', 'R2:C', 'R3:C'])

    @pytest.mark.asyncio
    async def test_close_pipeline(self):
        # m1:g1 ----------> \
        # m2:g2 -> m3:g2 -> m4:g3
        builder = ParallelPipeline()
        builder.add_module(ModuleTestTeardown('m1', group='g1'))
        builder.add_module(ModuleTestTeardown('m2', group='g2'))
        builder.add_module(
            ModuleTestTeardown('m3', group='g2')
            .depends_on(builder.get_module('m2'))
        )
        builder.add_module(
            ModuleTestTeardown('m4', group='g3')
            .depends_on(builder.get_module('m3'))
            .depends_on(builder.get_module('m1'))
            .expose_result('final')
        )

        manager = Manager()
        ns = manager.Namespace()
        ns.counter = 0

        pipeline = builder.build(dict(ns=ns))
        assert(ns.counter == len(pipeline.modules))

        await pipeline.run()
        await pipeline.close()

        assert pipeline.closed
        assert(ns.counter == 0)
        with pytest.raises(ClosedPipelineException):
            await pipeline.run()
        with pytest.raises(ClosedPipelineException):
            await pipeline.process()


class TestStatefulParallelPipeline:

    @expose()
    @finalize
    class ModuleA(Module.Runtime):
        def run(self, *args, **kwargs):
            return 'a'

    @expose()
    @accept(ModuleA, self=True)
    @finalize
    class ModuleB(Module.Runtime):
        def run(self, *args, **kwargs):
            return 'b'

    @expose('agg')
    @accept(ModuleA, ModuleB)
    @finalize
    class ModuleAgg(Module.Aggregate):
        def aggregate(self, data, **kwargs):
            """ Request Aggregate """
            self.add_data([(d.name, d.result) for d in data.collection])
            return self.state

    @expose()
    @accept(ModuleB, ModuleAgg, self=True)
    @finalize
    class ModuleC(Module.Runtime):
        def run(self, data, *args, **kwargs):
            if data.has('agg'):
                return data.get('agg') * 2
            else:
                return 'c'

    @pytest.mark.asyncio
    async def test_simple_pipeline_multiple_run_and_agg(self, ray_context):
        # m1:g1 ----------> \
        # m2:g2 -> m3:g2 -> m4:g3(agg)
        builder = ParallelPipeline()
        builder.add_module(self.ModuleA('m1', group='g1'))
        builder.add_module(self.ModuleB('m2', group='g2'))
        builder.add_module(
            self.ModuleB('m3', group='g2')
            .depends_on(builder.get_module('m2'))
        )
        builder.add_module(
            self.ModuleAgg('m4', group='g3')
            .depends_on(builder.get_module('m3'))
            .depends_on(builder.get_module('m1'))
            .expose_result('final')
        )

        pipeline = builder.build()
        r1 = await pipeline.run('R1')
        r2 = await pipeline.run('R2')
        agg1 = await pipeline.process('A1')
        r3 = await pipeline.run('R3')
        agg2 = await pipeline.process('A1')

        assert len(r1) == len(r2) == len(r3) == 3
        assert len(agg1['final']) == 2
        assert len(agg2['final']) == 1

    @pytest.mark.asyncio
    async def test_stateful_pipeline_mixed(self, ray_context):
        # m1:g1 -> m2:g1 -> \
        # m3:g2 -> m4:g2 ---> m5:g3 -> m6:g3(agg) -> m7:g3
        builder = ParallelPipeline()
        builder.add_module(self.ModuleA('m1', group='g1'))
        builder.add_module(self.ModuleB('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleA('m3', group='g2'))
        builder.add_module(self.ModuleB('m4', group='g2').depends_on(builder.get_module('m3')))
        builder.add_module(
            self.ModuleB('m5', group='g3')
            .depends_on(builder.get_module('m2'))
            .depends_on(builder.get_module('m4'))
        )
        builder.add_module(self.ModuleAgg('m6', group='g3').depends_on(builder.get_module('m5')))
        builder.add_module(self.ModuleC('m7', group='g3').depends_on(builder.get_module('m6')))

        pipeline = builder.build()
        res = await pipeline.run('R1')
        agg = await pipeline.process('A1')

        assert len(res) == 5
        assert len(agg['agg']) == 1
        assert len(agg['m7']) == 2

    @pytest.mark.asyncio
    async def test_stateful_pipeline_with_stateful_group(self, ray_context):
        # m1:g1 -> m2:g1 -> \
        # m3:g2 -> m4:g2 ---> m5:g3 -> m6:g4(agg) -> m7:g4
        builder = ParallelPipeline()
        builder.add_module(self.ModuleA('m1', group='g1'))
        builder.add_module(self.ModuleB('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleA('m3', group='g2'))
        builder.add_module(self.ModuleB('m4', group='g2').depends_on(builder.get_module('m3')))
        builder.add_module(
            self.ModuleB('m5', group='g3')
            .depends_on(builder.get_module('m2'))
            .depends_on(builder.get_module('m4'))
        )
        builder.add_module(self.ModuleAgg('m6', group='g4').depends_on(builder.get_module('m5')))
        builder.add_module(self.ModuleC('m7', group='g4').depends_on(builder.get_module('m6')))

        pipeline = builder.build()
        await pipeline.run('R1')
        res = await pipeline.run('R2')
        agg = await pipeline.process('A1')

        assert len(res) == 5
        assert len(agg) == 2
        assert len(agg['agg']) == 2
        assert agg['agg'][0][0] == agg['agg'][1][0] == ('m5', 'b')

    @pytest.mark.asyncio
    async def test_stateful_pipeline_with_two_afteragg_modules(self, ray_context):
        # m1:g1 -> m2:g1 -> \
        # m3:g2 -> m4:g2 ---> m5:g3 -> m6:g3(agg) -> m7:g4 -> m8:g4
        builder = ParallelPipeline()
        builder.add_module(self.ModuleA('m1', group='g1'))
        builder.add_module(self.ModuleB('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleA('m3', group='g2'))
        builder.add_module(self.ModuleB('m4', group='g2').depends_on(builder.get_module('m3')))
        builder.add_module(
            self.ModuleB('m5', group='g3')
            .depends_on(builder.get_module('m2'))
            .depends_on(builder.get_module('m4'))
        )
        builder.add_module(self.ModuleAgg('m6', group='g3').depends_on(builder.get_module('m5')))
        builder.add_module(self.ModuleC('m7', group='g4').depends_on(builder.get_module('m6')))
        builder.add_module(self.ModuleC('m8', group='g4').depends_on(builder.get_module('m7')))

        pipeline = builder.build()
        await pipeline.run('R1')
        await pipeline.run('R2')
        agg = await pipeline.process('A1')

        assert len(agg) == 3
        assert type(agg['agg']) == list
        assert len(agg['agg']) == 2
        assert type(agg['m7']) == list
        assert len(agg['m7']) == 4
        assert type(agg['m8']) == str
        assert agg['m8'] == 'c'

    @pytest.mark.asyncio
    async def test_stateful_pipeline_confluence_example(self, ray_context):
        # m1:g1 -> m2:g1 -> \   /--> m5:g3 -> m6:g3
        # m3:g2 ----------> m4:g3 -> agg7:g3 ----> m8:g4 -> m9:g4
        builder = ParallelPipeline()
        builder.add_module(self.ModuleA('m1', group='g1'))
        builder.add_module(self.ModuleB('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleA('m3', group='g2'))
        builder.add_module(
            self.ModuleB('m4', group='g3')
            .depends_on(builder.get_module('m2'))
            .depends_on(builder.get_module('m3'))
        )
        builder.add_module(self.ModuleB('m5', group='g3').depends_on(builder.get_module('m4')))
        builder.add_module(self.ModuleB('m6', group='g3').depends_on(builder.get_module('m5')))
        builder.add_module(self.ModuleAgg('agg7', group='g3').depends_on(builder.get_module('m4')))
        builder.add_module(self.ModuleC('m8', group='g4').depends_on(builder.get_module('agg7')))
        builder.add_module(self.ModuleC('m9', group='g4').depends_on(builder.get_module('m8')))

        pipeline = builder.build()
        await pipeline.run('R1')
        await pipeline.run('R2')
        res = await pipeline.run('R3')
        agg = await pipeline.process('A1')

        assert len(res) == 6
        assert len(agg) == 3
        assert type(agg['agg']) == list
        assert len(agg['agg']) == 3
        assert type(agg['m8']) == list
        assert len(agg['m8']) == 6
        assert type(agg['m9']) == str
        assert agg['m9'] == 'c'

    @pytest.mark.asyncio
    async def test_stateful_pipeline_with_two_agg_modules_double_run_and_agg(self, ray_context):
        #                                           /-> m12:g6 -> m13:g6
        # m1:g1 -> m2:g1 -> \   /--> m5:g3 -> m6:g3 --> agg10:g5 -> m11:g5
        # m3:g2 ----------> m4:g3 -> agg7:g3 ---------> m8:g4 -> m9:g4

        builder = ParallelPipeline()
        builder.add_module(self.ModuleA('m1', group='g1'))
        builder.add_module(self.ModuleB('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleA('m3', group='g2'))
        builder.add_module(
            self.ModuleB('m4', group='g3')
            .depends_on(builder.get_module('m2'))
            .depends_on(builder.get_module('m3'))
        )
        builder.add_module(self.ModuleB('m5', group='g3').depends_on(builder.get_module('m4')))
        builder.add_module(self.ModuleB('m6', group='g3').depends_on(builder.get_module('m5')))
        builder.add_module(self.ModuleAgg('agg7', group='g3').depends_on(builder.get_module('m4')))
        builder.add_module(self.ModuleC('m8', group='g4').depends_on(builder.get_module('agg7')))
        builder.add_module(self.ModuleC('m9', group='g4').depends_on(builder.get_module('m8')))
        builder.add_module(
            self.ModuleAgg('agg10', group='g5')
            .depends_on(builder.get_module('m6'))
            .expose_result('second_agg')
        )
        builder.add_module(self.ModuleC('m11', group='g5').depends_on(builder.get_module('agg10')))
        builder.add_module(self.ModuleB('m12', group='g6').depends_on(builder.get_module('m6')))
        builder.add_module(self.ModuleB('m13', group='g6').depends_on(builder.get_module('m12')))

        pipeline = builder.build()
        res1 = await pipeline.run('R1')
        agg1 = await pipeline.process('A1')
        res2 = await pipeline.run('R2')
        agg2 = await pipeline.process('A2')

        assert len(res1) == len(res2) == 8
        assert len(agg1) == len(agg2) == 5
        assert type(agg2['agg']) == list
        assert type(agg2['m8']) == list
        assert type(agg2['m9']) == str
        assert type(agg2['second_agg']) == list
        assert type(agg2['m11']) == str

    @pytest.mark.asyncio
    async def test_stateful_pipeline_invalid(self, ray_context):
        # m1:g1 -> m2:g1(agg)-\
        # m3:g2 -> m4:g2 ------> m5:g3
        builder = ParallelPipeline()
        builder.add_module(self.ModuleA('m1', group='g1'))
        builder.add_module(self.ModuleAgg('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleA('m3', group='g2'))
        builder.add_module(self.ModuleB('m4', group='g2').depends_on(builder.get_module('m3')))
        builder.add_module(
            self.ModuleC('m5', group='g3')
            .depends_on(builder.get_module('m2'))
            .depends_on(builder.get_module('m4'))
        )

        with pytest.raises(Exception):
            pipeline = builder.build()

    @pytest.mark.asyncio
    async def test_stateful_pipeline_with_only_nonregular_modules(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(self.ModuleAgg('m1', group='g1'))
        builder.add_module(self.ModuleC('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleC('m3', group='g2').depends_on(builder.get_module('m1')))

        pipeline = builder.build()
        res = await pipeline.run('R1')
        agg = await pipeline.process('A1')

        assert len(res) == 0
        assert len(agg) == 3
        assert agg['agg'] == [[]]
        assert agg['m2'] == agg['m3'] == [[], []]

    @pytest.mark.asyncio
    async def test_stateful_pipeline_with_nonregular_modules_different_groups(self, ray_context):
        builder = ParallelPipeline()
        builder.add_module(self.ModuleAgg('m1', group='g1'))
        builder.add_module(self.ModuleC('m2', group='g1').depends_on(builder.get_module('m1')))
        builder.add_module(self.ModuleC('m3', group='g2').depends_on(builder.get_module('m2')))

        pipeline = builder.build()
        res = await pipeline.run('R1')
        agg = await pipeline.process('A1')

        assert len(res) == 0
        assert len(agg) == 3
        assert agg['agg'] == [[]]
        assert agg['m2'] == [[], []]
        assert agg['m3'] == 'c'



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

dependencies = [
    'pyyaml>=5.3.1,<6',
    'ray>=1.2.0,<2',
]

test_dependencies = [
    'pytest==6.1.*',
    'pytest-cov==2.10.*',
    'pytest-asyncio==0.14.*',
]

setuptools.setup(
    name='magda',
    version='0.1.0',
    author='NeuroSYS Research',
    description=(
        'Library for building Modular and Asynchronous Graphs '
        'with Directed and Acyclic edges (MAGDA)'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/NeuroSYS-pl/magda',
    keywords=(
        'magda modular asynchronous parallelization'
        'directed-graphs acyclic-graphs pipeline python'
    ),
    packages=setuptools.find_packages(),
    license="Apache 2.0",
    entry_points={},
    python_requires='>=3.7',
    install_requires=dependencies,
    extras_require={
        'test': test_dependencies,
    },
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=test_dependencies,
)



from __future__ import annotations

from magda.module.module import Module


def module_typeguard(module_ref):
    if not issubclass(module_ref, Module):
        if issubclass(module_ref, Module.Runtime):
            raise TypeError(
                'Cannot decorate Module.Runtime class. '
                'Check decorators order: the @finalize '
                'must be first from the bottom!'
            )
        else:
            raise TypeError('The decorated class is not a Module!')



from time import sleep

from magda.module import Module
from magda.decorators import register, accept, finalize, produce

from examples.modules.common import logger, log
from examples.interfaces.common import Context
from examples.interfaces.fn import LambdaInterface


@accept(LambdaInterface)
@produce(LambdaInterface)
@register('B')
@finalize
class ModuleB(Module.Runtime):
    SLEEP_TIME = 1

    def bootstrap(self):
        ctx: Context = self.context
        log(self, ctx.timer, '--- Created!')

    def teardown(self):
        ctx: Context = self.context
        log(self, ctx.timer, '--- Teardown!')

    @logger
    def run(self, data: Module.ResultSet, *args, **kwargs):
        # Access strings (results) from the previous modules
        src = [text.fn() for text in data.of(LambdaInterface)]

        # Add delay for example purposes
        sleep(self.SLEEP_TIME)

        # Build output string and produce declared interface
        msg = '(' + ' + '.join(src) + (' = ' if len(src) else '') + f'{self.name})'
        return LambdaInterface(lambda: msg)



from __future__ import annotations

import asyncio
from typing import List

from magda.pipeline.parallel.group.actor import Actor


class ParallelActorPool:
    def __init__(self, actors: List[Actor]):
        self._replicas = len(actors)
        self._idle_actors = asyncio.Queue()
        for actor in actors:
            self._idle_actors.put_nowait(actor)

    @property
    def replicas(self) -> int:
        return self._replicas

    async def run(self, **job):
        actor = await self._idle_actors.get()
        result = await actor.run.remote(**job)
        await self._idle_actors.put(actor)
        return result

    async def teardown(self, **job):
        for _ in range(self._replicas):
            actor = await self._idle_actors.get()
            await actor.teardown.remote()



import pytest

from magda.module.factory import ModuleFactory
from magda.module.module import Module
from magda.decorators.finalizing import finalize


@finalize
class MockModule(Module.Runtime):
    pass


class TestFactory:

    def setup_method(self, method):
        ModuleFactory.unregister()

    def teardown_method(self, method):
        ModuleFactory.unregister()

    def test_factory_empty(self):
        assert len(ModuleFactory.module_references) == 0

    def test_factory_register(self):
        assert len(ModuleFactory.module_references) == 0

        ModuleFactory.register('MockModule', MockModule)
        assert len(ModuleFactory.module_references) == 1

    def test_factory_register_not_submodule(self):
        class NotSubmoduleMock:
            pass

        module_to_register = 'NotSubmoduleMock'
        with pytest.raises(TypeError):
            ModuleFactory.register(module_to_register, NotSubmoduleMock)

    def test_factory_register_same_module_twice(self):
        assert len(ModuleFactory.module_references) == 0

        module_to_register = 'SameModuleName'
        ModuleFactory.register(module_to_register, MockModule)
        assert len(ModuleFactory.module_references) == 1

        ModuleFactory.register(module_to_register, MockModule)
        assert len(ModuleFactory.module_references) == 1

    def test_factory_register_module_decorator(self):
        from test.mocks.module_decorated_mock import ModuleDecoratedMock
        assert len(ModuleFactory.module_references) == 1

    def test_factory_register_module_decorator_not_submodule(self):
        with pytest.raises(TypeError):
            @ModuleFactory.register('NotSubmoduleMock')
            class NotSubmoduleMock:
                pass

    def test_factory_create(self):
        module_name = 'MockModuleName'
        ModuleFactory.register(module_name, MockModule)
        MockModule.__abstractmethods__ = set()

        module_obj = ModuleFactory.create('Some name', module_name)
        assert type(module_obj) == MockModule

    def test_factory_create_module_type_not_registered(self):
        module_name = 'NotRegisteredModuleName'

        with pytest.raises(KeyError):
            ModuleFactory.create('Some name', module_name)

    def test_remove_registered_module(self):
        assert len(ModuleFactory.module_references) == 0
        ModuleFactory.register('MockModule', MockModule)
        assert len(ModuleFactory.module_references) == 1
        ModuleFactory.unregister('MockModule')
        assert len(ModuleFactory.module_references) == 0

    def test_remove_nonregistered_module(self):
        with pytest.warns(UserWarning):
            ModuleFactory.unregister('NotRegisteredModuleName')

    def test_nonstring_tag(self):
        with pytest.raises(TypeError):
            ModuleFactory.register(MockModule, MockModule)



from typing import List
from magda.module.module import Module
from magda.decorators.registering import register
from magda.decorators.finalizing import finalize


@register('ModuleDecoratedMock')
@finalize
class ModuleDecoratedMock(Module.Runtime):
    def run(self, data: List[Module.Result], **kwargs):
        return 'ModuleDecoratedMock'



import pytest

from magda.pipeline.sequential import SequentialPipeline
from magda.module.factory import ModuleFactory
from magda.module.module import Module
from magda.decorators.accepting import accept
from magda.decorators.finalizing import finalize
from magda.decorators.exposing import expose


@expose('a')
@accept(self=True)
@finalize
class MockModuleA(Module.Runtime):
    def run(self, **kwargs):
        pass


@accept(MockModuleA, self=True)
@finalize
class MockModuleB(Module.Runtime):
    def run(self, **kwargs):
        pass


class TestDecorator:

    def setup_method(self, method):
        ModuleFactory.unregister()
        ModuleFactory.register('MockModuleA', MockModuleA)
        ModuleFactory.register('MockModuleB', MockModuleB)

    def teardown_method(self, method):
        ModuleFactory.unregister()

    def test_accept_nonmodule(self):
        class SimpleClass:
            pass

        with pytest.raises(TypeError):
            @accept(SimpleClass, self=True)
            @finalize
            class InvalidModule(Module.Runtime):
                def run(self, **kwargs):
                    pass

    def test_module_without_finalize(self):
        with pytest.raises(TypeError):
            @accept()
            class InvalidModule(Module.Runtime):
                def run(self, **kwargs):
                    pass

    def test_module_with_wrong_decorator_order(self):
        with pytest.raises(TypeError):
            @finalize
            @expose('m1')
            class InvalidModule(Module.Runtime):
                def run(self, **kwargs):
                    pass

    def test_decorating_nonmodule_class(self):
        with pytest.raises(TypeError):
            @accept()
            class InvalidModule():
                def run(self, **kwargs):
                    pass

    def test_finalizing_nonmodule_class(self):
        with pytest.raises(TypeError):
            @finalize
            class NonModuleClass():
                pass

    def test_exposing_under_the_same_name(self):

        @expose('a')
        @accept(MockModuleA)
        @finalize
        class MockModuleC(Module.Runtime):
            def run(self, **kwargs):
                pass

        ModuleFactory.register('MockModuleC', MockModuleC)
        builder = SequentialPipeline()
        builder.add_module(MockModuleA('m1'))
        builder.add_module(MockModuleC('m2'))

        with pytest.raises(Exception):
            pipeline = builder.build()



import pytest

from magda.pipeline.sequential import SequentialPipeline
from magda.module.module import Module
from magda.decorators.accepting import accept
from magda.decorators.finalizing import finalize
from magda.exceptions import DisjointGraphException, CyclicDependenciesException


@accept(self=True)
@finalize
class SkipModule(Module.Runtime):
    def run(self, *args, **kwargs):
        return 'test module'


class TestGraph:
    def test_should_fail_on_not_connected_graph(self):
        builder = SequentialPipeline()
        builder.add_module(SkipModule('A'))
        builder.add_module(
            SkipModule('B')
            .depends_on(builder.get_module('A'))
        )
        builder.add_module(
            SkipModule('C')
            .depends_on(builder.get_module('A'))
            .depends_on(builder.get_module('B'))
        )
        builder.add_module(
            SkipModule('D')
        )
        with pytest.raises(DisjointGraphException):
            results = builder.build().run()

    def test_should_not_fail_on_graph_without_cycle(self):
        builder = SequentialPipeline()
        builder.add_module(SkipModule('A'))
        builder.add_module(
            SkipModule('B')
            .depends_on(builder.get_module('A'))
        )
        builder.add_module(
            SkipModule('C')
            .depends_on(builder.get_module('B'))
        )
        builder.add_module(
            SkipModule('D')
            .depends_on(builder.get_module('C'))
        )
        results = builder.build().run()
        assert 4 == len(builder.modules)

    def test_should_not_fail_on_branched_graph_without_cycle(self):
        builder = SequentialPipeline()
        builder.add_module(SkipModule('A'))
        builder.add_module(
            SkipModule('B')
            .depends_on(builder.get_module('A'))
        )
        builder.add_module(
            SkipModule('C')
            .depends_on(builder.get_module('A'))
        )
        builder.add_module(
            SkipModule('D')
            .depends_on(builder.get_module('B'))
            .depends_on(builder.get_module('C'))
        )

        results = builder.build().run()
        assert 4 == len(builder.modules)

    def test_should_fail_on_simple_cycle(self):
        builder = SequentialPipeline()
        module_a = SkipModule('A')
        module_b = SkipModule('B')
        module_a.depends_on(module_b)
        module_b.depends_on(module_a)
        builder.add_module(module_a)
        builder.add_module(module_b)

        with pytest.raises(CyclicDependenciesException):
            results = builder.build().run()

    def test_should_fail_on_long_cycle(self):
        builder = SequentialPipeline()
        module_a = SkipModule('A')
        module_side_a = SkipModule('side_A')
        module_b = SkipModule('B')
        module_c = SkipModule('C')
        module_d = SkipModule('D')
        module_side_d = SkipModule('side_D')
        module_e = SkipModule('E')
        module_f = SkipModule('F')
        module_g = SkipModule('G')
        module_side_g = SkipModule('side_G')

        # circular dependency
        module_b.depends_on(module_a)
        module_c.depends_on(module_b)
        module_d.depends_on(module_c)
        module_e.depends_on(module_d)
        module_f.depends_on(module_e)
        module_g.depends_on(module_f)
        module_a.depends_on(module_g)
        module_side_a.depends_on(module_a)
        module_side_d.depends_on(module_d)
        module_side_g.depends_on(module_g)

        builder.add_module(module_a)
        builder.add_module(module_b)
        builder.add_module(module_c)
        builder.add_module(module_d)
        builder.add_module(module_e)
        builder.add_module(module_f)
        builder.add_module(module_g)
        builder.add_module(module_side_a)
        builder.add_module(module_side_d)
        builder.add_module(module_side_g)

        with pytest.raises(CyclicDependenciesException):
            results = builder.build().run()

    def test_should_fail_on_multiple_cycles(self):
        builder = SequentialPipeline()
        module_a = SkipModule('A')
        module_side_a = SkipModule('side_A')
        module_b = SkipModule('B')
        module_c = SkipModule('C')
        module_d = SkipModule('D')
        module_side_d = SkipModule('side_D')
        module_e = SkipModule('E')
        module_f = SkipModule('F')
        module_g = SkipModule('G')
        module_side_g = SkipModule('side_G')

        # circular dependency with 4 additional side cycles
        module_b.depends_on(module_a)
        module_c.depends_on(module_b)
        module_d.depends_on(module_c)
        module_e.depends_on(module_d)
        module_f.depends_on(module_e)
        module_g.depends_on(module_f)
        module_a.depends_on(module_g)
        # additional cycle 1
        module_side_a.depends_on(module_a)
        module_c.depends_on(module_side_a)
        # additional cycle 2
        module_side_d.depends_on(module_d)
        module_g.depends_on(module_side_d)
        # additional cycle 3
        module_side_g.depends_on(module_g)
        module_b.depends_on(module_side_g)
        # additional cycle 4
        module_g.depends_on(module_side_g)

        builder.add_module(module_a)
        builder.add_module(module_b)
        builder.add_module(module_c)
        builder.add_module(module_d)
        builder.add_module(module_e)
        builder.add_module(module_f)
        builder.add_module(module_g)
        builder.add_module(module_side_a)
        builder.add_module(module_side_d)
        builder.add_module(module_side_g)

        with pytest.raises(CyclicDependenciesException):
            results = builder.build().run()
